(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/drivers/ip.js                                                //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Future = Npm.require("fibers/future");                             // 17
var Net = Npm.require("net");                                          // 18
var Dgram = Npm.require("dgram");                                      // 19
var Promise = Npm.require("es6-promise").Promise;                      // 20
var Capnp = Npm.require("capnp");                                      // 21
                                                                       //
var IpRpc = Capnp.importSystem("sandstorm/ip.capnp");                  // 23
                                                                       //
ByteStreamConnection = (function () {                                  // 25
  function ByteStreamConnection(connection) {                          // 26
    babelHelpers.classCallCheck(this, ByteStreamConnection);           //
                                                                       //
    this.connection = connection;                                      // 27
  }                                                                    //
                                                                       //
  ByteStreamConnection.prototype.done = (function () {                 // 25
    function done() {                                                  // 30
      this.connection.end();                                           // 31
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ByteStreamConnection.prototype.write = (function () {                // 25
    function write(data) {                                             // 34
      this.connection.write(data);                                     // 35
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  // expectSize not implemented                                        //
  // expectSize(size) { }                                              //
  return ByteStreamConnection;                                         //
})();                                                                  //
                                                                       //
IpInterfaceImpl = (function () {                                       // 42
  function IpInterfaceImpl(persistentMethods) {                        // 43
    babelHelpers.classCallCheck(this, IpInterfaceImpl);                //
                                                                       //
    _.extend(this, persistentMethods);                                 // 44
  }                                                                    //
                                                                       //
  IpInterfaceImpl.prototype.listenTcp = (function () {                 // 42
    function listenTcp(portNum, port) {                                // 47
      return new Promise(function (resolve, reject) {                  // 48
        var resolved = false;                                          // 49
        var server = Net.createServer(function (connection) {          // 50
          var wrappedConnection = new ByteStreamConnection(connection);
          var upstream = port.connect(wrappedConnection).upstream;     // 52
                                                                       //
          connection.on("data", function (data) {                      // 54
            upstream.write(data);                                      // 55
          });                                                          //
                                                                       //
          connection.on("close", function (hadError) {                 // 58
            upstream.done();                                           // 59
          });                                                          //
                                                                       //
          connection.on("error", function (err) {                      // 62
            if (resolved) {                                            // 63
              connection.write = errorWrite;                           // 64
            } else {                                                   //
              // upstream hasn't been resolved yet, so it's safe to reject
              reject(err);                                             // 67
            }                                                          //
          });                                                          //
        });                                                            //
                                                                       //
        server.listen(portNum, function () {                           // 72
          resolved = true;                                             // 73
          resolve({ handle: server }); // server has a close method which is all we want from a handle
        });                                                            //
                                                                       //
        server.on("error", function (err) {                            // 77
          if (!resolved) {                                             // 78
            reject(err);                                               // 79
          }                                                            //
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return listenTcp;                                                  //
  })();                                                                //
                                                                       //
  IpInterfaceImpl.prototype.listenUdp = (function () {                 // 42
    function listenUdp(portNum, port) {                                // 85
      return new Promise(function (resolve, reject) {                  // 86
        var portMap = {};                                              // 87
        var resolved = false;                                          // 88
        var server = Dgram.createSocket("udp4"); // TODO(someday): handle ipv6 sockets too
        server.bind(portNum);                                          // 90
                                                                       //
        server.on("listening", function () {                           // 92
          // Although UDP is connectionless, we don't resolve until here so that we can handle bind
          // errors such as invalid host                               //
          resolved = true;                                             // 95
          resolve({ handle: server }); // server has a close method which is all we want from a handle
        });                                                            //
                                                                       //
        server.on("error", function (err) {                            // 99
          // TODO(someday): do something about errors after the promise is resolved
          if (!resolved) {                                             // 101
            reject(err);                                               // 102
          } else {                                                     //
            console.error("error in listenUdp: " + err);               // 104
          }                                                            //
        });                                                            //
                                                                       //
        var returnMap = {};                                            // 108
        server.on("message", function (msg, rinfo) {                   // 109
          var address = rinfo.address + "]:" + rinfo.port;             // 110
          var returnPort = returnMap[address];                         // 111
                                                                       //
          if (!returnPort) {                                           // 113
            returnMap[address] = returnPort = new BoundUdpPortImpl(server, rinfo.address, rinfo.port);
          }                                                            //
                                                                       //
          port.send(msg, returnPort);                                  // 117
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return listenUdp;                                                  //
  })();                                                                //
                                                                       //
  return IpInterfaceImpl;                                              //
})();                                                                  //
                                                                       //
makeIpInterface = function (persistentMethods) {                       // 123
  return new Capnp.Capability(new IpInterfaceImpl(persistentMethods), IpRpc.PersistentIpInterface);
};                                                                     //
                                                                       //
BoundUdpPortImpl = (function () {                                      // 127
  function BoundUdpPortImpl(server, address, port) {                   // 128
    babelHelpers.classCallCheck(this, BoundUdpPortImpl);               //
                                                                       //
    this.server = server;                                              // 129
    this.address = address;                                            // 130
    this.port = port;                                                  // 131
  }                                                                    //
                                                                       //
  BoundUdpPortImpl.prototype.send = (function () {                     // 127
    function send(message, returnPort) {                               // 134
      // TODO(someday): this whole class is a hack to deal with the fact that we can't compare
      // capabilities or build a map with them. What we should be doing is mapping all ports to
      // their raw physical address/port, and using that here          //
      this.server.send(message, 0, message.length, this.port, this.address);
    }                                                                  //
                                                                       //
    return send;                                                       //
  })();                                                                //
                                                                       //
  return BoundUdpPortImpl;                                             //
})();                                                                  //
                                                                       //
var bits16 = Bignum(1).shiftLeft(16).sub(1);                           // 142
var bits32 = Bignum(1).shiftLeft(32).sub(1);                           // 143
                                                                       //
var intToIpv4 = function (num) {                                       // 145
  var part1 = num & 255;                                               // 146
  var part2 = num >> 8 & 255;                                          // 147
  var part3 = num >> 16 & 255;                                         // 148
  var part4 = num >> 24 & 255;                                         // 149
                                                                       //
  return part4 + "." + part3 + "." + part2 + "." + part1;              // 151
};                                                                     //
                                                                       //
var addressToString = function (address) {                             // 154
  var ipv6num = Bignum(address.upper64).shiftLeft(64).add(Bignum(address.lower64));
                                                                       //
  if (ipv6num.shiftRight(32).eq(bits16)) {                             // 157
    // this is an ipv4 address, we should return it as such            //
    var ipv4num = ipv6num.and(bits32).toNumber();                      // 159
    return intToIpv4(ipv4num);                                         // 160
  }                                                                    //
                                                                       //
  var hex = ipv6num.toString(16);                                      // 163
  var numColons = 0;                                                   // 164
  var out = "";                                                        // 165
                                                                       //
  for (var i = 0; i < hex.length; ++i) {                               // 167
    // start with lower bits of address and build the output in reverse
    // this ensures that we can place a colon every 4 characters       //
    out += hex[hex.length - 1 - i];                                    // 170
    if ((i + 1) % 4 === 0) {                                           // 171
      out += ":";                                                      // 172
      ++numColons;                                                     // 173
    }                                                                  //
  }                                                                    //
                                                                       //
  // Double colon represents all bits being 0                          //
  if (numColons < 7) {                                                 // 178
    out += "::";                                                       // 179
  }                                                                    //
                                                                       //
  return out.split("").reverse().join("");                             // 182
};                                                                     //
                                                                       //
var addressType = function (address) {                                 // 185
  var type = "udp4";                                                   // 186
                                                                       //
  // Check if it's an ipv6 address                                     //
  // TODO(someday): make this less hacky and change address to explicitly pass this information
  if (address.indexOf(":") != -1) {                                    // 190
    type = "udp6";                                                     // 191
  }                                                                    //
                                                                       //
  return type;                                                         // 194
};                                                                     //
                                                                       //
IpNetworkImpl = (function () {                                         // 197
  function IpNetworkImpl(persistentMethods) {                          // 198
    babelHelpers.classCallCheck(this, IpNetworkImpl);                  //
                                                                       //
    _.extend(this, persistentMethods);                                 // 199
  }                                                                    //
                                                                       //
  IpNetworkImpl.prototype.getRemoteHost = (function () {               // 197
    function getRemoteHost(address) {                                  // 202
      return { host: new IpRemoteHostImpl(address) };                  // 203
    }                                                                  //
                                                                       //
    return getRemoteHost;                                              //
  })();                                                                //
                                                                       //
  IpNetworkImpl.prototype.getRemoteHostByName = (function () {         // 197
    function getRemoteHostByName(address) {                            // 206
      return { host: new IpRemoteHostImpl(address) };                  // 207
    }                                                                  //
                                                                       //
    return getRemoteHostByName;                                        //
  })();                                                                //
                                                                       //
  return IpNetworkImpl;                                                //
})();                                                                  //
                                                                       //
makeIpNetwork = function (persistentMethods) {                         // 211
  return new Capnp.Capability(new IpNetworkImpl(persistentMethods), IpRpc.PersistentIpNetwork);
};                                                                     //
                                                                       //
IpRemoteHostImpl = (function () {                                      // 215
  function IpRemoteHostImpl(address) {                                 // 216
    babelHelpers.classCallCheck(this, IpRemoteHostImpl);               //
                                                                       //
    if (address.upper64 || address.upper64 === 0) {                    // 217
      // address is an ip.capnp:IpAddress, we need to convert it       //
      this.address = addressToString(address);                         // 219
    } else {                                                           //
      this.address = address;                                          // 221
    }                                                                  //
  }                                                                    //
                                                                       //
  IpRemoteHostImpl.prototype.getTcpPort = (function () {               // 215
    function getTcpPort(portNum) {                                     // 225
      return { port: new TcpPortImpl(this.address, portNum) };         // 226
    }                                                                  //
                                                                       //
    return getTcpPort;                                                 //
  })();                                                                //
                                                                       //
  IpRemoteHostImpl.prototype.getUdpPort = (function () {               // 215
    function getUdpPort(portNum) {                                     // 229
      return { port: new UdpPortImpl(this.address, portNum) };         // 230
    }                                                                  //
                                                                       //
    return getUdpPort;                                                 //
  })();                                                                //
                                                                       //
  return IpRemoteHostImpl;                                             //
})();                                                                  //
                                                                       //
TcpPortImpl = (function () {                                           // 234
  function TcpPortImpl(address, portNum) {                             // 235
    babelHelpers.classCallCheck(this, TcpPortImpl);                    //
                                                                       //
    this.address = address;                                            // 236
    this.port = portNum;                                               // 237
  }                                                                    //
                                                                       //
  TcpPortImpl.prototype.connect = (function () {                       // 234
    function connect(downstream) {                                     // 240
      var _this = this;                                                // 241
      var resolved = false;                                            // 242
      return new Promise(function (resolve, reject) {                  // 243
        var client = Net.connect({ host: _this.address, port: _this.port }, function () {
          resolved = true;                                             // 245
          resolve({ upstream: new ByteStreamConnection(client) });     // 246
        });                                                            //
                                                                       //
        client.on("data", function (data) {                            // 249
          downstream.write(data);                                      // 250
        });                                                            //
                                                                       //
        client.on("close", function (hadError) {                       // 253
          downstream.done();                                           // 254
        });                                                            //
                                                                       //
        client.on("error", function (err) {                            // 257
          if (resolved) {                                              // 258
            client.write = errorWrite;                                 // 259
          } else {                                                     //
            // upstream hasn't been resolved yet, so it's safe to reject
            reject(err);                                               // 262
          }                                                            //
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return connect;                                                    //
  })();                                                                //
                                                                       //
  return TcpPortImpl;                                                  //
})();                                                                  //
                                                                       //
var errorWrite = function (data) {                                     // 269
  throw new Error("error occurred in connection");                     // 270
};                                                                     //
                                                                       //
UdpPortImpl = (function () {                                           // 273
  function UdpPortImpl(address, portNum) {                             // 274
    babelHelpers.classCallCheck(this, UdpPortImpl);                    //
                                                                       //
    this.address = address;                                            // 275
    this.port = portNum;                                               // 276
                                                                       //
    var type = addressType(address);                                   // 278
    this.socket = Dgram.createSocket(type);                            // 279
                                                                       //
    // TODO(someday): close socket after a certain time of inactivity?
    // This may be pointless since grains are killed frequently when not in use anyways
                                                                       //
    // Temporary hack. We only expect clients to pass in a single return port, so we'll store it
    // and only send replies here.                                     //
    // This will be changed to be correct when equality comparisons are added to capabilities.
    this.returnPort = null;                                            // 287
                                                                       //
    var _this = this;                                                  // 289
    this.socket.on("message", function (msg, rinfo) {                  // 290
      if (_this.returnPort) {                                          // 291
        _this.returnPort.send(msg, _this);                             // 292
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  UdpPortImpl.prototype.send = (function () {                          // 273
    function send(message, returnPort) {                               // 297
      this.returnPort = returnPort;                                    // 298
      this.socket.send(message, 0, message.length, this.port, this.address);
                                                                       //
      // TODO(someday): use callback to catch errors and do something with them
    }                                                                  //
                                                                       //
    return send;                                                       //
  })();                                                                //
                                                                       //
  return UdpPortImpl;                                                  //
})();                                                                  //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=ip.js.map
