(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/grain-server.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
                                                                       //
var emailLinkWithInlineStyle = function (url, text) {                  // 19
  return "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + text + "</a>";
};                                                                     //
                                                                       //
Meteor.publish("grainTopBar", function (grainId) {                     // 26
  check(grainId, String);                                              // 27
                                                                       //
  var result = [Grains.find({                                          // 29
    _id: grainId,                                                      // 31
    $or: [{ userId: this.userId }, { "private": { $ne: true } }]       // 32
  }, {                                                                 //
    fields: {                                                          // 37
      title: 1,                                                        // 38
      userId: 1,                                                       // 39
      identityId: 1,                                                   // 40
      "private": 1                                                     // 41
    }                                                                  //
  })];                                                                 //
  if (this.userId) {                                                   // 45
    var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
    result.push(ApiTokens.find({                                       // 47
      grainId: grainId,                                                // 48
      $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }]
    }));                                                               //
  }                                                                    //
                                                                       //
  return result;                                                       // 56
});                                                                    //
                                                                       //
// We allow users to learn package information about a grain they own.
// This is used for obtaining icon and app title information for grains
// you own, which is used in the sidebar. It is not a security/privacy
// risk since it only exposes this information for grains the user owns.
Meteor.publish("packageByGrainId", function (grainId) {                // 63
  check(grainId, String);                                              // 64
  var publishThis = [];                                                // 65
  // We need to publish the packageId so that client-side code can     //
  // find the right package.                                           //
  var thisGrainCursor = Grains.find({                                  // 68
    _id: grainId,                                                      // 69
    userId: this.userId                                                // 70
  }, {                                                                 //
    fields: { packageId: 1 }                                           // 72
  });                                                                  //
  publishThis.push(thisGrainCursor);                                   // 74
                                                                       //
  if (thisGrainCursor.count()) {                                       // 76
    var thisGrain = thisGrainCursor.fetch()[0];                        // 77
    var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
    publishThis.push(thisPackageCursor);                               // 79
  }                                                                    //
                                                                       //
  return publishThis;                                                  // 82
});                                                                    //
                                                                       //
Meteor.publish("tokenInfo", function (token) {                         // 85
  // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
  // metadata that it will need to display the app icon and title.  We do not care about making
  // the metadata reactive.                                            //
  check(token, String);                                                // 89
                                                                       //
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  var apiToken = ApiTokens.findOne({                                   // 92
    _id: hashedToken                                                   // 93
  }, {                                                                 //
    fields: {                                                          // 95
      grainId: 1,                                                      // 96
      owner: 1,                                                        // 97
      revoked: 1                                                       // 98
    }                                                                  //
  });                                                                  //
  if (!apiToken) {                                                     // 101
    this.added("tokenInfo", token, { invalidToken: true });            // 102
  } else if (apiToken.revoked) {                                       //
    this.added("tokenInfo", token, { revoked: true });                 // 104
  } else {                                                             //
    var grainId = apiToken.grainId;                                    // 106
    var grain = Grains.findOne({                                       // 107
      _id: grainId                                                     // 108
    }, {                                                               //
      fields: {                                                        // 110
        packageId: 1,                                                  // 111
        appId: 1                                                       // 112
      }                                                                //
    });                                                                //
    if (!grain) {                                                      // 115
      this.added("tokenInfo", token, { invalidToken: true });          // 116
    } else {                                                           //
      if (apiToken.owner && apiToken.owner.user) {                     // 118
        var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
        var metadata = apiToken.owner.user.denormalizedGrainMetadata;  // 120
        if (identity && metadata) {                                    // 121
          this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
            grainId: grainId,                                          // 124
            grainMetadata: metadata });                                // 125
        } else {                                                       //
          this.added("tokenInfo", token, { invalidToken: true });      // 127
        }                                                              //
      } else if (!apiToken.owner || "webkey" in apiToken.owner) {      //
        if (this.userId) {                                             // 130
          var user = Meteor.users.findOne({ _id: this.userId });       // 131
          var identityIds = SandstormDb.getUserIdentityIds(user);      // 132
          var childToken = ApiTokens.findOne({ "owner.user.identityId": { $in: identityIds },
            parentToken: apiToken._id });                              // 134
          if (childToken) {                                            // 135
            this.added("tokenInfo", token, { alreadyRedeemed: true, grainId: apiToken.grainId });
            this.ready();                                              // 137
            return;                                                    // 138
          }                                                            //
        }                                                              //
                                                                       //
        var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
        var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
        var appIcon = undefined;                                       // 144
        if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
          var icons = pkg.manifest.metadata.icons;                     // 146
          appIcon = icons.grain || icons.appGrid;                      // 147
        }                                                              //
                                                                       //
        var denormalizedGrainMetadata = {                              // 150
          appTitle: appTitle,                                          // 151
          icon: appIcon,                                               // 152
          appId: appIcon ? undefined : grain.appId                     // 153
        };                                                             //
        this.added("tokenInfo", token, { webkey: true,                 // 155
          grainId: grainId,                                            // 157
          grainMetadata: denormalizedGrainMetadata });                 // 158
      } else {                                                         //
        this.added("tokenInfo", token, { invalidToken: true });        // 160
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  this.ready();                                                        // 165
  return;                                                              // 166
});                                                                    //
                                                                       //
Meteor.publish("requestingAccess", function (grainId) {                // 169
  check(grainId, String);                                              // 170
                                                                       //
  if (!this.userId) {                                                  // 172
    throw new Meteor.Error(403, "Must be logged in to request access.");
  }                                                                    //
                                                                       //
  var identityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: this.userId }));
                                                                       //
  var grain = globalDb.getGrain(grainId);                              // 178
  if (!grain) {                                                        // 179
    throw new Meteor.Error(404, "Grain not found.");                   // 180
  }                                                                    //
                                                                       //
  if (grain.userId === this.userId) {                                  // 183
    this.added("grantedAccessRequests", Random.id(), { grainId: grainId, identityId: grain.identityId });
  }                                                                    //
                                                                       //
  var _this = this;                                                    // 188
  var query = ApiTokens.find({ grainId: grainId, accountId: grain.userId,
    parentToken: { $exists: false },                                   // 190
    "owner.user.identityId": { $in: identityIds },                     // 191
    revoked: { $ne: true } });                                         // 192
  var handle = query.observe({                                         // 193
    added: function (apiToken) {                                       // 194
      _this.added("grantedAccessRequests", Random.id(), { grainId: grainId, identityId: apiToken.owner.user.identityId });
    }                                                                  //
  });                                                                  //
                                                                       //
  this.onStop(function () {                                            // 200
    return handle.stop();                                              //
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.publish("grainSize", function (grainId) {                       // 203
  // Publish pseudo-collection containing the size of the grain opened in the given session.
  check(grainId, String);                                              // 205
                                                                       //
  var grain = Grains.findOne(grainId);                                 // 207
  if (!grain || grain.userId !== this.userId) {                        // 208
    return [];                                                         // 209
  }                                                                    //
                                                                       //
  var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
  var _this = this;                                                    // 214
  var stopped = false;                                                 // 215
  var promise = getGrainSize(supervisor);                              // 216
                                                                       //
  function getNext(oldSize) {                                          // 218
    promise = getGrainSize(supervisor, oldSize);                       // 219
    promise.then(function (size) {                                     // 220
      if (!stopped) {                                                  // 221
        if (size !== oldSize) {                                        // 222
          // sometimes there are false alarms                          //
          _this.changed("grainSizes", grainId, { size: size });        // 223
        }                                                              //
                                                                       //
        getNext(size);                                                 // 226
      }                                                                //
    }, function (err) {                                                //
                                                                       //
      if (!stopped) {                                                  // 230
        if (err.kjType === "disconnected") {                           // 231
          _this.stop();                                                // 232
        } else {                                                       //
          _this.error(err);                                            // 234
        }                                                              //
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  promise.then(function (size) {                                       // 240
    if (!stopped) {                                                    // 241
      _this.added("grainSizes", grainId, { size: size });              // 242
      _this.ready();                                                   // 243
      getNext(size);                                                   // 244
    }                                                                  //
  }, function (err) {                                                  //
                                                                       //
    if (!stopped) {                                                    // 248
      if (err.kjType === "disconnected") {                             // 249
        _this.stop();                                                  // 250
      } else {                                                         //
        _this.error(err);                                              // 252
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  _this.onStop(function () {                                           // 257
    stopped = true;                                                    // 258
    promise.cancel();                                                  // 259
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.methods({                                                       // 263
  deleteGrain: function (grainId) {                                    // 264
    check(grainId, String);                                            // 265
                                                                       //
    if (this.userId) {                                                 // 267
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 269
        Grains.remove(grainId);                                        // 270
        globalDb.removeApiTokens({                                     // 271
          grainId: grainId,                                            // 272
          $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }]
        });                                                            //
        if (grain.lastUsed) {                                          // 278
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 280
        }                                                              //
                                                                       //
        if (!this.isSimulation) {                                      // 283
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 285
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  forgetGrain: function (grainId, identityId) {                        // 291
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 295
    check(identityId, String);                                         // 296
                                                                       //
    if (!this.userId) {                                                // 298
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
                                                                       //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      globalDb.removeApiTokens({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
                                                                       //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 307
    check(grainId, String);                                            // 308
    check(identityId, String);                                         // 309
    if (!this.userId) {                                                // 310
      throw new Meteor.Error(403, "Must be logged in.");               // 311
    }                                                                  //
                                                                       //
    var grain = globalDb.getGrain(grainId) || {};                      // 314
    if (!grain.userId === this.userId) {                               // 315
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
                                                                       //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
                                                                       //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 322
    check(grainId, String);                                            // 323
    check(newTitle, String);                                           // 324
    check(identityId, String);                                         // 325
    if (this.userId) {                                                 // 326
      var grain = Grains.findOne(grainId);                             // 327
      if (grain) {                                                     // 328
        if (grain.userId === this.userId) {                            // 329
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 332
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
                                                                       //
          var token = ApiTokens.findOne({                              // 336
            grainId: grainId,                                          // 337
            objectId: { $exists: false },                              // 338
            "owner.user.identityId": identityId                        // 339
          }, {                                                         //
            sort: { created: 1 }                                       // 341
          });                                                          //
          if (token) {                                                 // 343
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  privatizeGrain: function (grainId) {                                 // 351
    check(grainId, String);                                            // 352
    if (this.userId) {                                                 // 353
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
                                                                       //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    var _this2 = this;                                                 //
                                                                       //
    if (!this.isSimulation) {                                          // 360
      var _ret = (function () {                                        //
        check(origin, String);                                         // 361
        check(identityId, String);                                     // 362
        check(grainId, String);                                        // 363
        check(title, String);                                          // 364
        check(roleAssignment, roleAssignmentPattern);                  // 365
        check(contacts, [{                                             // 366
          _id: String,                                                 // 368
          isDefault: Match.Optional(Boolean),                          // 369
          profile: Match.ObjectIncluding({                             // 370
            service: String,                                           // 371
            name: String,                                              // 372
            intrinsicName: String                                      // 373
          })                                                           //
        }]);                                                           //
        check(message, { text: String, html: String });                // 377
        if (!_this2.userId) {                                          // 378
          throw new Meteor.Error(403, "Must be logged in to share by email.");
        }                                                              //
                                                                       //
        if (!globalDb.userHasIdentity(_this2.userId, identityId)) {    // 382
          throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
        }                                                              //
                                                                       //
        if (contacts.length === 0) {                                   // 386
          throw new Meteor.Error(400, "No contacts were provided.");   // 387
        }                                                              //
                                                                       //
        var accountId = _this2.userId;                                 // 390
        var identity = globalDb.getIdentity(identityId);               // 391
        var sharerDisplayName = identity.profile.name;                 // 392
        var outerResult = { successes: [], failures: [] };             // 393
        contacts.forEach(function (contact) {                          // 394
          if (contact.isDefault && contact.profile.service === "email") {
            var emailAddress = contact.profile.intrinsicName;          // 396
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
            var url = origin + "/shared/" + result.token;              // 401
            var html = message.html + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
            try {                                                      // 407
              SandstormEmail.send({                                    // 408
                to: emailAddress,                                      // 409
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 410
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
                html: html                                             // 415
              });                                                      //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString() });
            }                                                          //
          } else {                                                     //
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
            var url = origin + "/shared/" + result.token;              // 425
            try {                                                      // 426
              var _identity = Meteor.users.findOne({ _id: contact._id });
              var email = _.findWhere(SandstormDb.getVerifiedEmails(_identity), { primary: true });
              if (email) {                                             // 430
                var intrinsicName = contact.profile.intrinsicName;     // 431
                var loginNote = undefined;                             // 432
                if (contact.profile.service === "google") {            // 433
                  loginNote = "Google account with address " + email.email;
                } else if (contact.profile.service === "github") {     //
                  loginNote = "Github account with username " + intrinsicName;
                } else if (contact.profile.service === "email") {      //
                  loginNote = "email address " + intrinsicName;        // 438
                } else {                                               //
                  throw new Meteor.Error(500, "Unknown service to email share link.");
                }                                                      //
                                                                       //
                var html = message.html + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
                SandstormEmail.send({                                  // 448
                  to: email.email,                                     // 449
                  from: "Sandstorm server <no-reply@" + HOSTNAME + ">",
                  subject: sharerDisplayName + " has invited you to join a grain: " + title,
                  text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                  html: html                                           // 455
                });                                                    //
              } else {                                                 //
                outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + url + " with them." });
              }                                                        //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString(),
                warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
            }                                                          //
          }                                                            //
        });                                                            //
                                                                       //
        return {                                                       // 470
          v: outerResult                                               //
        };                                                             //
      })();                                                            //
                                                                       //
      if (typeof _ret === "object") return _ret.v;                     //
    }                                                                  //
  },                                                                   //
                                                                       //
  requestAccess: function (origin, grainId, identityId) {              // 474
    check(origin, String);                                             // 475
    check(grainId, String);                                            // 476
    check(identityId, String);                                         // 477
    if (!this.isSimulation) {                                          // 478
      if (!this.userId) {                                              // 479
        throw new Meteor.Error(403, "Must be logged in to request access.");
      }                                                                //
                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 483
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
                                                                       //
      var grain = Grains.findOne(grainId);                             // 487
      if (!grain) {                                                    // 488
        throw new Meteor.Error(404, "No such grain");                  // 489
      }                                                                //
                                                                       //
      var grainOwner = globalDb.getUser(grain.userId);                 // 492
      var email = _.findWhere(SandstormDb.getUserEmails(grainOwner), { primary: true });
      if (!email) {                                                    // 494
        throw new Meteor.Error("no email", "Grain owner has no email address.");
      }                                                                //
                                                                       //
      var emailAddress = email.email;                                  // 498
                                                                       //
      var identity = globalDb.getIdentity(identityId);                 // 500
      globalDb.addContact(grainOwner._id, identityId);                 // 501
                                                                       //
      var envelopeFrom = globalDb.getReturnAddress();                  // 503
      var fromEmail = globalDb.getServerTitle() + " <" + globalDb.getReturnAddress() + ">";
      var senderEmails = SandstormDb.getVerifiedEmails(identity);      // 505
      var senderPrimaryEmail = _.findWhere(senderEmails, { primary: true });
      var accountPrimaryEmailAddress = Meteor.user().primaryEmail;     // 507
      if (_.findWhere(senderEmails, { email: accountPrimaryEmailAddress })) {
        fromEmail = accountPrimaryEmailAddress;                        // 509
      } else if (senderPrimaryEmail) {                                 //
        fromEmail = senderPrimaryEmail.email;                          // 511
      }                                                                //
                                                                       //
      // TODO(soon): In the HTML version, we should display an identity card.
      var identityNote = "";                                           // 515
      if (identity.profile.service === "google") {                     // 516
        identityNote = " (" + identity.privateIntrinsicName + ")";     // 517
      } else if (identity.profile.service === "github") {              //
        identityNote = " (" + identity.profile.intrinsicName + " on GitHub)";
      } else if (identity.profile.service === "email") {               //
        identityNote = " (" + identity.profile.intrinsicName + ")";    // 521
      }                                                                //
                                                                       //
      var message = identity.profile.name + identityNote + " is requesting access to your grain: " + grain.title + ".";
                                                                       //
      var url = origin + "/share/" + grainId + "/" + identityId;       // 527
                                                                       //
      var html = message + "<br><br>" + emailLinkWithInlineStyle(url, "Open Sharing Menu");
                                                                       //
      var user = Meteor.user();                                        // 532
      var ACCESS_REQUEST_LIMIT = 10;                                   // 533
      var resetCount = true;                                           // 534
      if (user.accessRequests) {                                       // 535
        if (user.accessRequests.resetOn < new Date()) {                // 536
          Meteor.users.update({ _id: user._id }, { $unset: { accessRequests: 1 } });
        } else if (user.accessRequests.count >= ACCESS_REQUEST_LIMIT) {
          throw new Meteor.Error(403, "For spam control reasons, you are not allowed to make " + "more than " + ACCESS_REQUEST_LIMIT + " access requests per day.");
        } else {                                                       //
          resetCount = false;                                          // 543
        }                                                              //
      }                                                                //
                                                                       //
      var modifier = { $inc: { "accessRequests.count": 1 } };          // 547
      if (resetCount) {                                                // 548
        var tomorrow = new Date(Date.now() + 86400000);                // 549
        modifier.$set = { "accessRequests.resetOn": tomorrow };        // 550
      }                                                                //
                                                                       //
      Meteor.users.update({ _id: user._id }, modifier);                // 553
                                                                       //
      SandstormEmail.send({                                            // 555
        to: emailAddress,                                              // 556
        envelopeFrom: envelopeFrom,                                    // 557
        from: fromEmail,                                               // 558
        subject: grain.title + " - Request for access",                // 559
        text: message + "\n\nFollow this link to share access:\n\n" + url,
        html: html                                                     // 561
      });                                                              //
    }                                                                  //
  }                                                                    //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain-server.js.map
