(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/proxy.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var ChildProcess = Npm.require("child_process");                       // 18
var Fs = Npm.require("fs");                                            // 19
var Path = Npm.require("path");                                        // 20
var Future = Npm.require("fibers/future");                             // 21
var Http = Npm.require("http");                                        // 22
var Url = Npm.require("url");                                          // 23
var Promise = Npm.require("es6-promise").Promise;                      // 24
var Capnp = Npm.require("capnp");                                      // 25
var Net = Npm.require("net");                                          // 26
                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");  // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;   // 33
                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).
  check(accountId, String);                                            // 38
  check(identityId, String);                                           // 39
  check(apiTokenAccountId, String);                                    // 40
                                                                       //
  // Bail out early if quota enforcement is disabled.                  //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 43
    return;                                                            // 44
  }                                                                    //
                                                                       //
  var aliceAccountId = apiTokenAccountId;                              // 47
  var bobAccountId = accountId;                                        // 48
  var bobIdentityId = identityId;                                      // 49
                                                                       //
  if (Meteor.users.find({                                              // 51
    _id: bobAccountId,                                                 // 52
    referredByComplete: { $exists: true }                              // 53
  }).count() > 0) {                                                    //
    return;                                                            // 55
  }                                                                    //
                                                                       //
  // Only actually update Bob's Identity ID if there is no referredBy.
  Meteor.users.update({ _id: bobIdentityId, referredBy: { $exists: false } }, { $set: { referredBy: aliceAccountId } });
};                                                                     //
                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {         // 64
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share
  // access' interface. Let's call them Bob.                           //
  //                                                                   //
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the
  // first referredBy we find -- let's call that Alice.                //
  //                                                                   //
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's
  // Identity since it has become redundant.                           //
  //                                                                   //
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a
  // link with Bob, even if Bob already had an account.                //
                                                                       //
  // Bail out early if quota support is not enabled.                   //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 80
    return;                                                            // 81
  }                                                                    //
                                                                       //
  // Bail out if Bob has a referredByComplete.                         //
  if (Meteor.users.find({ _id: bobAccountId, referredByComplete: { $exists: true } }).count() > 0) {
    return;                                                            // 86
  }                                                                    //
                                                                       //
  // Look for a referredBy on any of Bob's identities.                 //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: bobAccountId }));
  var bobIdentityWithReferredBy = Meteor.users.findOne({               // 91
    _id: { $in: bobIdentityIds },                                      // 92
    referredBy: { $exists: true }                                      // 93
  });                                                                  //
                                                                       //
  if (!bobIdentityWithReferredBy) {                                    // 96
    return;                                                            // 97
  }                                                                    //
                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy;           // 100
                                                                       //
  // Store Bob's Account.referralCompletedBy.                          //
  var now = new Date();                                                // 103
  Meteor.users.update({                                                // 104
    _id: bobAccountId,                                                 // 105
    referredByComplete: { $exists: false }                             // 106
  }, {                                                                 //
    $set: {                                                            // 108
      referredByComplete: bobIdentityWithReferredBy.referredBy,        // 109
      referredCompleteDate: now                                        // 110
    }                                                                  //
  });                                                                  //
                                                                       //
  // Update Alice's Account.referredIdentityIds.                       //
  Meteor.users.update({ _id: aliceAccountId }, {                       // 115
    $push: { referredIdentityIds: bobIdentityWithReferredBy._id }      // 116
  });                                                                  //
                                                                       //
  // Remove now-redundant Bob identity referredBy.                     //
  Meteor.users.update({ _id: bobIdentityWithReferredBy._id }, {        // 120
    $unset: { referredBy: true }                                       // 121
  });                                                                  //
}                                                                      //
                                                                       //
// User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double
// backslashes. For security reasons, we MUST NOT whitelist any user-agents
// that may render html and execute embedded scripts.                  //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");
                                                                       //
var SESSION_PROXY_TIMEOUT = 60000;                                     // 140
                                                                       //
var sandstormCoreFactory = makeSandstormCoreFactory();                 // 142
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
var sandstormBackend = sandstormBackendConnection.restore(null, Backend);
                                                                       //
// TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.
globalBackend = new SandstormBackend(globalDb, sandstormBackend);      // 149
Meteor.onConnection(function (connection) {                            // 150
  connection.sandstormBackend = globalBackend;                         // 151
});                                                                    //
                                                                       //
// We've observed a problem in production where occasionally the front-end stops talking to the
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.
// Restarting the node process (and only the node process) always fixes the problem.
//                                                                     //
// Here, I've added some code that attempts to detect the problem by doing a health check
// periodically and then remaking the connection if it seems broken. We'll see if this helps!
var backendHealthy = true;                                             // 161
var disconnectCount = 0;                                               // 162
Meteor.setInterval(function () {                                       // 163
  if (!backendHealthy) {                                               // 164
    if (disconnectCount++ > 2) process.abort();                        // 165
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);
    sandstormBackendConnection.close();                                // 168
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);
    globalBackend._backendCap = sandstormBackend;                      // 171
  }                                                                    //
                                                                       //
  var debugLog = !backendHealthy;                                      // 174
  backendHealthy = false;                                              // 175
  var promise = sandstormBackend.ping().then(function () {             // 176
    backendHealthy = true;                                             // 177
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);
  }, function (err) {                                                  //
    console.error("error: Backend ping threw error!", err.stack);      // 180
    // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.
    // So we don't expect this branch to execute in any case.          //
  });                                                                  //
                                                                       //
  if (debugLog) {                                                      // 186
    console.log("capnp.js: outer promise:", promise);                  // 187
  }                                                                    //
}, 30000);                                                             //
                                                                       //
// =======================================================================================
// Meteor context <-> Async Node.js context adapters                   //
// TODO(cleanup):  Move to a different file.                           //
                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {    // 195
  callback();                                                          // 196
});                                                                    //
                                                                       //
inMeteor = function (callback) {                                       // 199
  // Calls the callback in a Meteor context.  Returns a Promise for its result.
  return new Promise(function (resolve, reject) {                      // 201
    inMeteorInternal(function () {                                     // 202
      try {                                                            // 203
        resolve(callback());                                           // 204
      } catch (err) {                                                  //
        reject(err);                                                   // 206
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promiseToFuture = function (promise) {                                 // 212
  var result = new Future();                                           // 213
  promise.then(result["return"].bind(result), result["throw"].bind(result));
  return result;                                                       // 215
};                                                                     //
                                                                       //
waitPromise = function (promise) {                                     // 218
  return promiseToFuture(promise).wait();                              // 219
};                                                                     //
                                                                       //
// =======================================================================================
// API for creating / starting grains from Meteor methods.             //
                                                                       //
var proxiesByHostId = {};                                              // 225
                                                                       //
Meteor.methods({                                                       // 227
  newGrain: function (packageId, command, title, identityId) {         // 228
    // Create and start a new grain.                                   //
                                                                       //
    check(packageId, String);                                          // 231
    check(command, Object); // Manifest.Command from package.capnp.    // 232
    check(title, String);                                              // 233
    check(identityId, String);                                         // 234
                                                                       //
    if (!this.userId) {                                                // 236
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");
    }                                                                  //
                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {          // 240
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!isSignedUpOrDemo()) {                                         // 244
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");
    }                                                                  //
                                                                       //
    if (isUserOverQuota(Meteor.user())) {                              // 249
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 254
    var isDev = false;                                                 // 255
    if (!pkg) {                                                        // 256
      // Maybe they wanted a dev package.  Check there too.            //
      pkg = DevPackages.findOne(packageId);                            // 258
      isDev = true;                                                    // 259
    }                                                                  //
                                                                       //
    if (!pkg) {                                                        // 262
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");
    }                                                                  //
                                                                       //
    var appId = pkg.appId;                                             // 266
    var manifest = pkg.manifest;                                       // 267
    var grainId = Random.id(22); // 128 bits of entropy                // 268
    Grains.insert({                                                    // 269
      _id: grainId,                                                    // 270
      packageId: packageId,                                            // 271
      appId: appId,                                                    // 272
      appVersion: manifest.appVersion,                                 // 273
      userId: this.userId,                                             // 274
      identityId: identityId,                                          // 275
      title: title,                                                    // 276
      "private": true                                                  // 277
    });                                                                //
                                                                       //
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev);
    globalBackend.updateLastActive(grainId, this.userId, identityId);  // 281
    return grainId;                                                    // 282
  },                                                                   //
                                                                       //
  openSession: function (grainId, identityId, cachedSalt) {            // 285
    // Open a new UI session on an existing grain.  Starts the grain if it is not already
    // running.                                                        //
                                                                       //
    check(grainId, String);                                            // 289
    check(identityId, Match.OneOf(undefined, null, String));           // 290
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 291
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!Grains.findOne({ _id: grainId })) {                           // 297
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + grainId);
    }                                                                  //
                                                                       //
    var db = this.connection.sandstormDb;                              // 301
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 302
    if (!SandstormPermissions.mayOpenGrain(db, { grain: { _id: grainId, identityId: identityId } })) {
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
    }                                                                  //
                                                                       //
    var opened = globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt);
    var result = opened.methodResult;                                  // 310
    var proxy = new Proxy(grainId, this.userId, result.sessionId, result.hostId, result.tabId, identityId, false, opened.supervisor);
    proxiesByHostId[result.hostId] = proxy;                            // 313
    return result;                                                     // 314
  },                                                                   //
                                                                       //
  openSessionFromApiToken: function (params, identityId, cachedSalt) {
    // Given an API token, either opens a new WebSession to the underlying grain or returns a
    // path to which the client should redirect in order to open such a session.
                                                                       //
    check(params, {                                                    // 321
      token: String,                                                   // 322
      incognito: Boolean                                               // 323
    });                                                                //
    check(identityId, Match.OneOf(undefined, null, String));           // 325
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 326
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    var token = params.token;                                          // 332
    var incognito = params.incognito;                                  // 333
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne(hashedToken);                     // 335
    validateWebkey(apiToken);                                          // 336
    var grain = Grains.findOne({ _id: apiToken.grainId });             // 337
    if (!grain) {                                                      // 338
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);
    }                                                                  //
                                                                       //
    if (apiToken.accountId) {                                          // 342
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);
    }                                                                  //
                                                                       //
    var title = undefined;                                             // 346
    if (grain.userId === apiToken.accountId) {                         // 347
      title = grain.title;                                             // 348
    } else {                                                           //
      var sharerToken = apiToken.identityId && ApiTokens.findOne({     // 350
        grainId: apiToken.grainId,                                     // 352
        "owner.user.identityId": apiToken.identityId                   // 353
      }, {                                                             //
        sort: {                                                        // 355
          "owner.user.lastUsed": -1                                    // 356
        }                                                              //
      });                                                              //
      if (sharerToken) {                                               // 359
        title = sharerToken.owner.user.title;                          // 360
      } else {                                                         //
        title = "shared grain";                                        // 362
      }                                                                //
    }                                                                  //
                                                                       //
    if (this.userId && !incognito) {                                   // 366
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({ "owner.user.identityId": identityId, parentToken: hashedToken })) {
        var owner = { user: { identityId: identityId, title: title } };
                                                                       //
        // Create a new API token for the identity redeeming this token.
        var result = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: token }, apiToken.grainId, apiToken.petname, { allAccess: null }, owner);
        globalDb.addContact(apiToken.accountId, identityId);           // 374
                                                                       //
        // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For
        // share-by-identity, this happens immediately.)               //
        if (result.parentApiToken) {                                   // 380
          var parentApiToken = result.parentApiToken;                  // 381
          if (parentApiToken.forSharing && parentApiToken.accountId) {
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return { redirectToGrain: apiToken.grainId };                    // 389
    } else {                                                           //
      if (!SandstormPermissions.mayOpenGrain(globalDb, { token: apiToken })) {
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
      }                                                                //
                                                                       //
      var opened = globalBackend.openSessionInternal(apiToken.grainId, null, null, title, apiToken, cachedSalt);
                                                                       //
      var result = opened.methodResult;                                // 399
      var proxy = new Proxy(apiToken.grainId, grain.userId, result.sessionId, result.hostId, result.tabId, identityId, false);
      proxy.apiToken = apiToken;                                       // 402
      proxiesByHostId[result.hostId] = proxy;                          // 403
      return result;                                                   // 404
    }                                                                  //
  },                                                                   //
                                                                       //
  keepSessionAlive: function (sessionId) {                             // 408
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared
    //   by them.                                                      //
    check(sessionId, String);                                          // 411
                                                                       //
    var session = Sessions.findAndModify({                             // 413
      query: { _id: sessionId },                                       // 414
      update: { $set: { timestamp: new Date().getTime() } },           // 415
      fields: { grainId: 1, identityId: 1, hostId: 1 }                 // 416
    });                                                                //
                                                                       //
    if (session) {                                                     // 419
      // Session still present in database, so send keep-alive to backend.
      try {                                                            // 421
        var grainId = session.grainId;                                 // 422
        var hostId = session.hostId;                                   // 423
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;
        if (!supervisor) {                                             // 425
          supervisor = globalBackend.continueGrain(grainId).supervisor;
        }                                                              //
                                                                       //
        waitPromise(supervisor.keepAlive());                           // 429
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);
      } catch (err) {                                                  //
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on
        // the next request, so whatever.                              //
        if (err.kjType !== "disconnected") {                           // 434
          throw err;                                                   // 435
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 439
    } else {                                                           //
      return false;                                                    // 441
    }                                                                  //
  },                                                                   //
                                                                       //
  shutdownGrain: function (grainId) {                                  // 445
    check(grainId, String);                                            // 446
    var grain = Grains.findOne(grainId);                               // 447
    if (!grain || !this.userId || grain.userId !== this.userId) {      // 448
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");
    }                                                                  //
                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));
  }                                                                    //
});                                                                    //
                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {        // 456
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the
  // `expiresIfUnused` field is reset to `refreshedExpiration`.        //
                                                                       //
  if (!apiToken) {                                                     // 461
    throw new Meteor.Error(403, "Invalid authorization token");        // 462
  }                                                                    //
                                                                       //
  if (apiToken.revoked) {                                              // 465
    throw new Meteor.Error(403, "Authorization token has been revoked");
  }                                                                    //
                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {               // 469
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");
  }                                                                    //
                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {  // 473
    throw new Meteor.Error(403, "Authorization token expired");        // 474
  }                                                                    //
                                                                       //
  if (apiToken.expiresIfUnused) {                                      // 477
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {            // 478
      throw new Meteor.Error(403, "Authorization token expired");      // 479
    } else if (refreshedExpiration) {                                  //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: refreshedExpiration } });
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                     // 488
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");
  }                                                                    //
};                                                                     //
                                                                       //
// Used by shared/grain.js (which sounds like a broken dependency, since this code is only available
// on the server)                                                      //
getGrainSize = function (supervisor, oldSize) {                        // 495
  var promise = undefined;                                             // 496
  if (oldSize === undefined) {                                         // 497
    promise = supervisor.getGrainSize();                               // 498
  } else {                                                             //
    promise = supervisor.getGrainSizeWhenDifferent(oldSize);           // 500
  }                                                                    //
                                                                       //
  var promise2 = promise.then(function (result) {                      // 503
    return parseInt(result.size);                                      // 503
  });                                                                  //
  promise2.cancel = function () {                                      // 504
    promise.cancel();                                                  // 504
  };                                                                   //
                                                                       //
  return promise2;                                                     // 506
};                                                                     //
                                                                       //
Meteor.startup(function () {                                           // 509
  var shutdownApp = function (appId) {                                 // 510
    Grains.find({ appId: appId }).forEach(function (grain) {           // 511
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));
    });                                                                //
  };                                                                   //
                                                                       //
  DevPackages.find().observe({                                         // 516
    removed: function (devPackage) {                                   // 517
      shutdownApp(devPackage.appId);                                   // 517
    },                                                                 //
                                                                       //
    changed: function (oldDevPackage, newDevPackage) {                 // 519
      shutdownApp(oldDevPackage.appId);                                // 520
      if (oldDevPackage.appId !== newDevPackage.appId) {               // 521
        shutdownApp(newDevPackage.appId);                              // 522
      }                                                                //
    },                                                                 //
                                                                       //
    added: function (devPackage) {                                     // 526
      shutdownApp(devPackage.appId);                                   // 526
    }                                                                  //
  });                                                                  //
                                                                       //
  Sessions.find().observe({                                            // 529
    removed: function (session) {                                      // 530
      var proxy = proxiesByHostId[session.hostId];                     // 531
      delete proxiesByHostId[session.hostId];                          // 532
      if (proxy) {                                                     // 533
        proxy.close();                                                 // 534
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Kill off sessions idle for >~3 minutes.                             //
var TIMEOUT_MS = 180000;                                               // 541
var gcSessions = function () {                                         // 542
  var now = new Date().getTime();                                      // 543
  Sessions.remove({ timestamp: { $lt: now - TIMEOUT_MS } });           // 544
};                                                                     //
                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                   // 547
                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {           // 549
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it
  // doesn't already exist. The first request on the session host will always create a new proxy.
  // Later requests may create a proxy if they go to a different front-end replica or if the
  // front-end was restarted.                                          //
  check(hostId, String);                                               // 554
                                                                       //
  return Promise.resolve(undefined).then(function () {                 // 556
    var proxy = proxiesByHostId[hostId];                               // 557
    if (proxy) {                                                       // 558
      return proxy;                                                    // 559
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByHostId[hostId] = null;                                  // 562
                                                                       //
      return inMeteor(function () {                                    // 564
        var session = Sessions.findOne({ hostId: hostId });            // 565
        if (!session) {                                                // 566
          if (isAlreadyOpened) {                                       // 567
            return new Promise(function (resolve, reject) {            // 568
              var observer = undefined;                                // 569
              var task = Meteor.setTimeout(function () {               // 570
                observer.stop();                                       // 571
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                               //
              observer = Sessions.find({ hostId: hostId }).observe({   // 578
                added: function () {                                   // 579
                  observer.stop();                                     // 580
                  Meteor.clearTimeout(task);                           // 581
                  resolve(getProxyForHostId(hostId, false));           // 582
                }                                                      //
              });                                                      //
            });                                                        //
          } else {                                                     //
            // Does not appear to be a valid session host.             //
            return undefined;                                          // 588
          }                                                            //
        }                                                              //
                                                                       //
        var apiToken = undefined;                                      // 592
        if (session.hashedToken) {                                     // 593
          apiToken = ApiTokens.findOne({ _id: session.hashedToken });  // 594
          // We don't have to fully validate the API token here because if it changed the session
          // would have been deleted.                                  //
          if (!apiToken) {                                             // 597
            throw new Meteor.Error(410, "ApiToken has been deleted");  // 598
          }                                                            //
        }                                                              //
                                                                       //
        var grain = Grains.findOne(session.grainId);                   // 602
        if (!grain) {                                                  // 603
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 605
        }                                                              //
                                                                       //
        // Note that we don't need to call mayOpenGrain() because the existence of a session
        // implies this check was already performed.                   //
                                                                       //
        var proxy = new Proxy(grain._id, grain.userId, session._id, hostId, session.tabId, session.identityId, false);
        if (apiToken) proxy.apiToken = apiToken;                       // 613
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the user's access was revoked).                     //
        if (hostId in proxiesByHostId) {                               // 617
          proxiesByHostId[hostId] = proxy;                             // 618
        } else {                                                       //
          throw new Meteor.Error(403, "Session was concurrently closed.");
        }                                                              //
                                                                       //
        return proxy;                                                  // 623
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// API tokens                                                          //
                                                                       //
var ApiSessionProxies = (function () {                                 //
  // Class that caches proxies for requests that come in through HTTP API endpoints. Such
  // requests are not associated with any entry in the `Sessions` collection, so we need
  // some other means of detecting unused proxies and closing them. This class works by keeping
  // the proxies in two buckets. When a proxy is used, it is put in the new bucket. Periodically,
  // the old bucket is emptied and the contents of new bucket are moved to the old bucket.
  //                                                                   //
  // Each bucket is a two-level map, keyed by hashes of the tokens and their `ApiSession.Params`
  // structs.                                                          //
                                                                       //
  function ApiSessionProxies(intervalMillis) {                         // 642
    var _this2 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, ApiSessionProxies);              //
                                                                       //
    check(intervalMillis, Number);                                     // 643
    this.newBucket = {};                                               // 644
    this.oldBucket = {};                                               // 645
    this.interval = Meteor.setInterval(function () {                   // 646
      for (var oldHashedToken in babelHelpers.sanitizeForInObject(_this2.oldBucket)) {
        for (var oldHashedParams in babelHelpers.sanitizeForInObject(_this2.oldBucket[oldHashedToken])) {
          var newProxy = _this2.newBucket[oldHashedToken] && _this2.newBucket[oldHashedToken][oldHashedParams];
          var oldProxy = _this2.oldBucket[oldHashedToken][oldHashedParams];
          if (oldProxy && !newProxy) {                                 // 652
            if (Object.keys(oldProxy.websockets).length > 0) {         // 653
              // A client has an open websocket. Keep the proxy around.
              _this2.put(oldHashedToken, oldHashedParams, oldProxy);   // 655
            } else {                                                   //
              // We can close this proxy and forget about it.          //
              oldProxy.close();                                        // 658
            }                                                          //
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      _this2.oldBucket = _this2.newBucket;                             // 664
      _this2.newBucket = {};                                           // 665
    }, intervalMillis);                                                //
  }                                                                    //
                                                                       //
  ApiSessionProxies.prototype.get = (function () {                     // 632
    function get(hashedToken, hashedParams) {                          // 669
      if (this.newBucket[hashedToken] && this.newBucket[hashedToken][hashedParams]) {
        return this.newBucket[hashedToken][hashedParams];              // 671
      } else if (this.oldBucket[hashedToken] && this.oldBucket[hashedToken][hashedParams]) {
        var proxy = this.oldBucket[hashedToken][hashedParams];         // 673
        delete this.oldBucket[hashedToken][hashedParams];              // 674
        this.put(hashedToken, hashedParams, proxy);                    // 675
        return proxy;                                                  // 676
      } else {                                                         //
        return null;                                                   // 678
      }                                                                //
    }                                                                  //
                                                                       //
    return get;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.put = (function () {                     // 632
    function put(hashedToken, hashedParams, proxy) {                   // 682
      if (!this.newBucket[hashedToken]) {                              // 683
        this.newBucket[hashedToken] = {};                              // 684
      }                                                                //
                                                                       //
      this.newBucket[hashedToken][hashedParams] = proxy;               // 687
    }                                                                  //
                                                                       //
    return put;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.removeProxiesOfToken = (function () {    // 632
    function removeProxiesOfToken(hashedToken) {                       // 690
      if (hashedToken in this.oldBucket) {                             // 691
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.oldBucket[hashedToken])) {
          this.oldBucket[hashedToken][hashedParams].close();           // 693
        }                                                              //
                                                                       //
        delete this.oldBucket[hashedToken];                            // 696
      }                                                                //
                                                                       //
      if (hashedToken in this.newBucket) {                             // 699
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.newBucket[hashedToken])) {
          this.newBucket[hashedToken][hashedParams].close();           // 701
        }                                                              //
                                                                       //
        delete this.newBucket[hashedToken];                            // 704
      }                                                                //
    }                                                                  //
                                                                       //
    return removeProxiesOfToken;                                       //
  })();                                                                //
                                                                       //
  return ApiSessionProxies;                                            //
})();                                                                  //
                                                                       //
var apiSessionProxies = new ApiSessionProxies(3 * 60 * 1000);          // 709
                                                                       //
Meteor.startup(function () {                                           // 711
  var clearSessionsAndProxies = function (token) {                     // 712
    // Clears all sessions and API proxies associated with `token` or any token that is downstream
    // in the sharing graph.                                           //
    // TODO(soon): Only clear sessions and proxies for which the permissions have changed.
    var downstream = SandstormPermissions.downstreamTokens(globalDb, { token: token });
    downstream.push(token);                                            // 717
    var identityIds = [];                                              // 718
    var tokenIds = [];                                                 // 719
                                                                       //
    downstream.forEach(function (token) {                              // 721
      apiSessionProxies.removeProxiesOfToken(token._id);               // 722
                                                                       //
      tokenIds.push(token._id);                                        // 724
      if (token.owner && token.owner.user) {                           // 725
        identityIds.push(token.owner.user.identityId);                 // 726
      }                                                                //
    });                                                                //
                                                                       //
    Sessions.find({                                                    // 730
      grainId: token.grainId,                                          // 731
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    }, {                                                               //
      fields: { hostId: 1 }                                            // 736
    }).forEach(function (session) {                                    //
      var proxy = proxiesByHostId[session.hostId];                     // 738
      if (proxy) {                                                     // 739
        proxy.close();                                                 // 740
      }                                                                //
                                                                       //
      delete proxiesByHostId[session.hostId];                          // 743
    });                                                                //
                                                                       //
    Sessions.remove({                                                  // 746
      grainId: token.grainId,                                          // 747
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    });                                                                //
  };                                                                   //
                                                                       //
  Grains.find().observe({                                              // 755
    changed: function (newGrain, oldGrain) {                           // 756
      if (oldGrain["private"] != newGrain["private"]) {                // 757
        Sessions.remove({ grainId: oldGrain._id, identityId: { $ne: oldGrain.identityId } });
        ApiTokens.find({ grainId: oldGrain._id }).forEach(function (apiToken) {
          apiSessionProxies.removeProxiesOfToken(apiToken._id);        // 760
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  ApiTokens.find({ grainId: { $exists: true }, objectId: { $exists: false } }).observe({
    added: function (newApiToken) {                                    // 767
      // TODO(soon): Unfortunately, added() gets called for all existing role assignments when the
      //   front-end restarts, meaning clearing sessions here will cause people's views to refresh
      //   on server upgrade, which is not a nice user experience. It's also sad to force-refresh
      //   people when they gained new permissions since they might be in the middle of something,
      //   and it's not strictly necessary for security. OTOH, it's sad to be non-reactive. Maybe
      //   we should notify people that they have new permissions and let them click a thing to
      //   refresh?                                                    //
      //clearSessions(roleAssignment.grainId, roleAssignment.recipient);
      //clearApiProxies(roleAssignment.grainId);                       //
    },                                                                 //
                                                                       //
    changed: function (newApiToken, oldApiToken) {                     // 779
      if (!_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
        clearSessionsAndProxies(newApiToken);                          // 782
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldApiToken) {                                  // 786
      clearSessionsAndProxies(oldApiToken);                            // 787
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
function getApiSessionParams(request) {                                // 792
  var params = {};                                                     // 793
  if ("x-sandstorm-passthrough" in request.headers) {                  // 794
    var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {
      return s.trim();                                                 // 797
    });                                                                //
    // The only currently supported passthrough value is 'address', but others could be useful in
    // the future.                                                     //
                                                                       //
    if (optIns.indexOf("address") !== -1) {                            // 801
      // Sadly, we can't use request.socket.remoteFamily because it's not available in the
      // (rather-old) version of node that comes in the Meteor bundle we're using. Hence this
      // hackery.                                                      //
      var addressToPass = request.socket.remoteAddress;                // 805
      if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {
        // Allow overriding the socket's remote address with X-Real-IP header if the request comes
        // from either localhost or an RFC1918 address. These are not useful for geolocation anyway.
        addressToPass = request.headers["x-real-ip"];                  // 809
      }                                                                //
                                                                       //
      if (Net.isIPv4(addressToPass)) {                                 // 812
        // Map IPv4 addresses in IPv6.                                 //
        // This conveniently comes out to a 48-bit number, which is precisely representable in a
        // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we
        // might otherwise need to precisely represent 64-bit fields.  //
        var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {
          return parseInt(x, 10);                                      // 818
        }).reduce(function (a, b) {                                    //
          return 256 * a + b;                                          // 819
        });                                                            //
        params.remoteAddress = {                                       // 820
          lower64: v4Int,                                              // 821
          upper64: 0                                                   // 822
        };                                                             //
      } else if (Net.isIPv6(addressToPass)) {                          //
        // TODO(test): Unit test this                                  //
        // Parse a valid v6 address.                                   //
        // Split into groups, then insert an appropriate number of 0's if :: was used.
        var groups = addressToPass.split(":");                         // 828
                                                                       //
        // Strip extra empty group in the case of a leading or trailing '::'.
        if (groups[0] === "") {                                        // 831
          groups.shift();                                              // 832
        }                                                              //
                                                                       //
        if (groups[groups.length - 1] === "") {                        // 835
          groups.pop();                                                // 836
        }                                                              //
                                                                       //
        var lastGroup = groups[groups.length - 1];                     // 839
        // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
        // should expand into two groups of 4-character hex strings, like the rest of the address.
        if (Net.isIPv4(lastGroup)) {                                   // 842
          groups.pop();                                                // 843
          var _quad = lastGroup.split(".").map(function (x) {          // 844
            return parseInt(x, 10);                                    // 844
          });                                                          //
          groups.push((_quad[0] * 256 + _quad[1]).toString(16));       // 845
          groups.push((_quad[2] * 256 + _quad[3]).toString(16));       // 846
        }                                                              //
                                                                       //
        var groupsToAdd = 8 - groups.length;                           // 849
        var emptyGroupIndex = groups.indexOf("");                      // 850
        var cleanGroups = undefined;                                   // 851
        if (emptyGroupIndex !== -1) {                                  // 852
          var head = groups.slice(0, emptyGroupIndex);                 // 853
          // groupsToAdd + 1 because we sliced out the empty element   //
          var mid = Array(groupsToAdd + 1);                            // 855
          for (var i = 0; i < groupsToAdd + 1; i++) {                  // 856
            mid[i] = "0";                                              // 857
          }                                                            //
                                                                       //
          var tail = groups.slice(emptyGroupIndex + 1, groups.length);
          cleanGroups = [].concat(head, mid, tail);                    // 861
        } else {                                                       //
          cleanGroups = groups;                                        // 863
        }                                                              //
                                                                       //
        var ints = cleanGroups.map(function (x) {                      // 866
          return parseInt(x, 16);                                      // 866
        });                                                            //
        // We use strings because we'd lose data from loss of precision casting the 64-bit uints
        // into 53-bit-mantissa doubles.                               //
        params.remoteAddress = {                                       // 869
          upper64: quadToIntString(ints.slice(0, 4)),                  // 870
          lower64: quadToIntString(ints.slice(4, 8))                   // 871
        };                                                             //
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return Capnp.serialize(ApiSession.Params, params);                   // 877
}                                                                      //
                                                                       //
// Used by server/drivers/external-ui/view.js                          //
getProxyForApiToken = function (token, request) {                      // 881
  check(token, String);                                                // 882
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  var tabId = Crypto.createHash("sha256").update("tab:").update(hashedToken).digest("hex").slice(0, 32);
  var serializedParams = getApiSessionParams(request);                 // 886
  var hashedParams = Crypto.createHash("sha256").update(serializedParams).digest("base64");
  return Promise.resolve(undefined).then(function () {                 // 888
    var proxy = apiSessionProxies.get(hashedToken, hashedParams);      // 889
    if (proxy) {                                                       // 890
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {    // 891
        throw new Meteor.Error(403, "Authorization token expired");    // 892
      }                                                                //
                                                                       //
      return proxy;                                                    // 895
    } else {                                                           //
      return inMeteor(function () {                                    // 897
        var tokenInfo = ApiTokens.findOne(hashedToken);                // 898
        validateWebkey(tokenInfo);                                     // 899
                                                                       //
        var grain = Grains.findOne(tokenInfo.grainId);                 // 901
        if (!grain) {                                                  // 902
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 904
        }                                                              //
                                                                       //
        var proxy = undefined;                                         // 907
        if (tokenInfo.userInfo) {                                      // 908
          throw new Error("API tokens created with arbitrary userInfo no longer supported");
        } else {                                                       //
          var identityId = null;                                       // 911
          if (tokenInfo.identityId && !tokenInfo.forSharing) {         // 912
            identityId = tokenInfo.identityId;                         // 913
          }                                                            //
                                                                       //
          proxy = new Proxy(tokenInfo.grainId, grain.userId, null, null, tabId, identityId, true);
          proxy.apiToken = tokenInfo;                                  // 917
          proxy.apiSessionParams = serializedParams;                   // 918
        }                                                              //
                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, { token: tokenInfo })) {
          // Note that only public grains may be opened without a user ID.
          throw new Meteor.Error(403, "Unauthorized.");                // 923
        }                                                              //
                                                                       //
        if (tokenInfo.expires) {                                       // 926
          proxy.expires = tokenInfo.expires;                           // 927
        }                                                              //
                                                                       //
        apiSessionProxies.put(hashedToken, hashedParams, proxy);       // 930
                                                                       //
        return proxy;                                                  // 932
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var apiUseBasicAuth = function (req, hostId) {                         // 938
  // If the request was to a token-specific host *and* the request has no Origin header (meaning
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we
  // want to allow basic auth at all is because a lot of existing client apps support only basic
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {
    return true;                                                       // 947
  }                                                                    //
                                                                       //
  // Historically, all API tokens were served from the same host rather than have token-specific
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud
  // client app.                                                       //
  //                                                                   //
  // Since many clients in the wild have already been configured to use the shared API host, we
  // must continue to support them, so this logic remains.             //
  var agent = req.headers["user-agent"];                               // 959
  return agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                    // 960
};                                                                     //
                                                                       //
var apiTokenForRequest = function (req, hostId) {                      // 963
  // Extract the API token from the request.                           //
                                                                       //
  var auth = req.headers.authorization;                                // 966
  var token = undefined;                                               // 967
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {          // 968
    token = auth.slice(7).trim();                                      // 969
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];
  } else {                                                             //
    token = undefined;                                                 // 974
  }                                                                    //
                                                                       //
  if (token && hostId !== "api") {                                     // 977
    // Verify that the token matches the specific host.                //
    if (hostId !== globalDb.apiHostIdForToken(token)) {                // 979
      token = undefined;                                               // 980
    }                                                                  //
  }                                                                    //
                                                                       //
  return token;                                                        // 984
};                                                                     //
                                                                       //
// =======================================================================================
// Routing to proxies.                                                 //
//                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyUpgrade = function (hostId, req, socket, head) {               // 992
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (globalDb.isApiHostId(hostId)) {                                  // 998
    var token = apiTokenForRequest(req, hostId);                       // 999
    if (token) {                                                       // 1000
      return getProxyForApiToken(token).then(function (proxy) {        // 1001
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1004
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1006
        return true;                                                   // 1007
      });                                                              //
    } else {                                                           //
      return Promise.resolve(false);                                   // 1010
    }                                                                  //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1015
        // Cross-origin requests are not allowed on UI session hosts.  //
        var origin = req.headers.origin;                               // 1017
        if (origin !== PROTOCOL + "//" + req.headers.host) {           // 1018
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);
        }                                                              //
                                                                       //
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1024
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1026
        return true;                                                   // 1027
      } else {                                                         //
        return false;                                                  // 1029
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyRequest = function (hostId, req, res) {                        // 1036
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  var hostIdHash = globalDb.isApiHostId(hostId);                       // 1042
  if (hostIdHash) {                                                    // 1043
    var _ret = (function () {                                          //
      // This is a request for the API host.                           //
                                                                       //
      res.setHeader("Access-Control-Allow-Origin", "*");               // 1046
                                                                       //
      if (req.method === "OPTIONS") {                                  // 1048
        // Reply to CORS preflight request.                            //
                                                                       //
        // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients
        // must send a valid Authorization header, and since cookies are not used for
        // authorization, this is perfectly safe. In a sane world, we would only need to send back
        // 'Access-Control-Allow-Origin: *' and be done with it.       //
        //                                                             //
        // However, CORS demands that we explicitly whitelist individual methods and headers for
        // use cross-origin, as if this is somehow useful for implementing any practical security
        // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one
        // individually. We cannot just write '*' for these lists. WTF, CORS?
        //                                                             //
        // Luckily, the request tells us exactly what method and headers are being requested, so we
        // only need to copy those over, rather than create an exhaustive list. But this is still
        // overly complicated.                                         //
                                                                       //
        var accessControlHeaders = {                                   // 1065
          "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",
          "Access-Control-Max-Age": "3600"                             // 1067
        };                                                             //
                                                                       //
        // Copy all requested headers to the allowed headers list.     //
        var requestedHeaders = req.headers["access-control-request-headers"];
        if (requestedHeaders) {                                        // 1072
          accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;
        }                                                              //
                                                                       //
        // Add the requested method to the allowed methods list, if it's not there already.
        var requestedMethod = req.headers["access-control-request-method"];
        if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {
          accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;
        }                                                              //
                                                                       //
        for (header in babelHelpers.sanitizeForInObject(accessControlHeaders)) {
          res.setHeader(header, accessControlHeaders[header]);         // 1084
        }                                                              //
      }                                                                //
                                                                       //
      var errorHandler = function (err) {                              // 1088
        if (err instanceof Meteor.Error) {                             // 1089
          console.log("error: " + err);                                // 1090
          res.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
        } else {                                                       //
          res.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
        }                                                              //
                                                                       //
        res.end(err.stack);                                            // 1096
      };                                                               //
                                                                       //
      var token = apiTokenForRequest(req, hostId);                     // 1099
      if (token && req.headers["x-sandstorm-token-keepalive"]) {       // 1100
        inMeteor(function () {                                         // 1101
          var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);
          check(keepaliveDuration, Match.Integer);                     // 1103
          var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
          validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));
        }).then(function () {                                          //
          res.writeHead(200, {});                                      // 1107
          res.end();                                                   // 1108
        }, errorHandler);                                              //
      } else if (token) {                                              //
        getProxyForApiToken(token, req).then(function (proxy) {        // 1111
          proxy.requestHandler(req, res);                              // 1112
        }, errorHandler);                                              //
      } else {                                                         //
        // No token. Look up static API host info.                     //
        inMeteor(function () {                                         // 1116
          var apiHost = globalDb.collections.apiHosts.findOne(hostIdHash);
                                                                       //
          if (req.method === "OPTIONS") {                              // 1119
            // OPTIONS request with no authorization token.            //
                                                                       //
            var dav = ((apiHost || {}).options || {}).dav || [];       // 1122
            if (dav.length > 0) {                                      // 1123
              res.setHeader("DAV", dav.join(", "));                    // 1124
              res.setHeader("Access-Control-Expose-Headers", "DAV");   // 1125
            }                                                          //
                                                                       //
            res.writeHead(200, {});                                    // 1128
            res.end();                                                 // 1129
          } else {                                                     //
            var resources = (apiHost || {}).resources || {};           // 1131
            var path = SandstormDb.escapeMongoKey(req.url.split("?")[0]);
            if (path in resources) {                                   // 1133
              // Serve a static resource.                              //
              var resource = resources[path];                          // 1135
              console.log(path, resources, resource);                  // 1136
              if (resource.language) res.setHeader("Content-Language", resource.language);
              if (resource.encoding) res.setHeader("Content-Encoding", resource.encoding);
              res.writeHead(200, {                                     // 1139
                "Content-Type": resource.type                          // 1140
              });                                                      //
              res.end(resource.body);                                  // 1142
            } else {                                                   //
              if (apiUseBasicAuth(req, hostId)) {                      // 1144
                res.writeHead(401, {                                   // 1145
                  "Content-Type": "text/plain",                        // 1146
                  "WWW-Authenticate": "Basic realm='Sandstorm API'"    // 1147
                });                                                    //
              } else {                                                 //
                // TODO(someday): Display some sort of nifty API browser.
                res.writeHead(403, { "Content-Type": "text/plain" });  // 1151
              }                                                        //
                                                                       //
              res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
            }                                                          //
          }                                                            //
        }).then(function () {}, errorHandler);                         //
      }                                                                //
                                                                       //
      return {                                                         // 1163
        v: Promise.resolve(true)                                       //
      };                                                               //
    })();                                                              //
                                                                       //
    if (typeof _ret === "object") return _ret.v;                       //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1167
        proxy.requestHandler(req, res);                                // 1168
        return true;                                                   // 1169
      } else {                                                         //
        return false;                                                  // 1171
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Proxy class                                                         //
//                                                                     //
// Connects to a grain and exports it on a wildcard host.              //
//                                                                     //
                                                                       //
var Proxy = (function () {                                             //
  function Proxy(grainId, ownerId, sessionId, hostId, tabId, identityId, isApi, supervisor) {
    var _this3 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, Proxy);                          //
                                                                       //
    this.grainId = grainId;                                            // 1185
    this.ownerId = ownerId;                                            // 1186
    this.identityId = identityId;                                      // 1187
    this.supervisor = supervisor; // note: optional parameter; we can reconnect
    this.sessionId = sessionId;                                        // 1189
    this.tabId = tabId;                                                // 1190
    this.isApi = isApi;                                                // 1191
    this.hasLoaded = false;                                            // 1192
    this.websockets = {};                                              // 1193
    this.websocketCounter = 0; // Used for generating unique socket IDs.
    if (sessionId) {                                                   // 1195
      if (!hostId) throw new Error("sessionId must come with hostId");
      if (isApi) throw new Error("API proxy shouldn't have sessionId");
      this.hostId = hostId;                                            // 1198
    } else {                                                           //
      if (!isApi) throw new Error("non-API proxy requires sessionId");
      if (hostId) throw new Error("API proxy sholudn't have hostId");  // 1201
    }                                                                  //
                                                                       //
    if (this.identityId) {                                             // 1204
      var identity = globalDb.getIdentity(this.identityId);            // 1205
      if (!identity) {                                                 // 1206
        throw new Error("identity not found: " + this.identityId);     // 1207
      }                                                                //
                                                                       //
      this.userInfo = {                                                // 1210
        displayName: { defaultText: identity.profile.name },           // 1211
        preferredHandle: identity.profile.handle,                      // 1212
        identityId: new Buffer(identity._id, "hex")                    // 1213
      };                                                               //
      if (identity.profile.pictureUrl) this.userInfo.pictureUrl = identity.profile.pictureUrl;
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;
    } else {                                                           //
      this.userInfo = {                                                // 1218
        displayName: { defaultText: "Anonymous User" },                // 1219
        preferredHandle: "anonymous"                                   // 1220
      };                                                               //
    }                                                                  //
                                                                       //
    var _this = this;                                                  // 1224
                                                                       //
    this.requestHandler = function (request, response) {               // 1226
      if (_this3.sessionId) {                                          // 1227
        // Implement /_sandstorm-init for setting the session cookie.  //
        var url = Url.parse(request.url, true);                        // 1229
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this.sessionId) {
          _this.doSessionInit(request, response, url.query.path);      // 1231
          return;                                                      // 1232
        }                                                              //
      }                                                                //
                                                                       //
      Promise.resolve(undefined).then(function () {                    // 1236
        var contentLength = request.headers["content-length"];         // 1237
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.
          return _this.handleRequestStreaming(request, response, contentLength, 0);
        } else {                                                       //
          return readAll(request).then(function (data) {               // 1243
            return _this.handleRequest(request, data, response, 0);    // 1244
          });                                                          //
        }                                                              //
      })["catch"](function (err) {                                     //
        _this.setHasLoaded();                                          // 1248
                                                                       //
        var body = err.stack;                                          // 1250
        if (err.cppFile) {                                             // 1251
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");
        }                                                              //
                                                                       //
        if (err.kjType) {                                              // 1255
          body += "\ntype: " + err.kjType;                             // 1256
        }                                                              //
                                                                       //
        if (response.headersSent) {                                    // 1259
          // Unfortunately, it's too late to tell the client what happened.
          console.error("HTTP request failed after response already sent:", body);
          response.end();                                              // 1262
        } else {                                                       //
          if (err instanceof Meteor.Error) {                           // 1264
            response.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
          } else {                                                     //
            response.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
          }                                                            //
                                                                       //
          response.end(body);                                          // 1270
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    this.upgradeHandler = function (request, socket, head) {           // 1275
      _this.handleWebSocket(request, socket, head, 0)["catch"](function (err) {
        console.error("WebSocket setup failed:", err.stack);           // 1277
        // TODO(cleanup):  Manually send back a 500 response?          //
        socket.destroy();                                              // 1279
      });                                                              //
    };                                                                 //
  }                                                                    //
                                                                       //
  Proxy.prototype.close = (function () {                               // 1183
    function close() {                                                 // 1284
      for (var socketIdx in babelHelpers.sanitizeForInObject(this.websockets)) {
        this.websockets[socketIdx].destroy();                          // 1286
      };                                                               //
                                                                       //
      this.websockets = {};                                            // 1289
                                                                       //
      if (this.session) {                                              // 1291
        this.session.close();                                          // 1292
        delete this.session;                                           // 1293
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1296
        this.uiView.close();                                           // 1297
        delete this.uiView;                                            // 1298
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1301
        this.supervisor.close();                                       // 1302
        delete this.supervisor;                                        // 1303
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  Proxy.prototype.getConnection = (function () {                       // 1183
    function getConnection() {                                         // 1307
      if (!this.supervisor) {                                          // 1308
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;
        this.uiView = null;                                            // 1310
      }                                                                //
                                                                       //
      if (!this.uiView) {                                              // 1313
        this.uiView = this.supervisor.getMainView().view;              // 1314
      }                                                                //
    }                                                                  //
                                                                       //
    return getConnection;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewWebSession = (function () {                  // 1183
    function _callNewWebSession(request, userInfo) {                   // 1318
      var params = Capnp.serialize(WebSession.Params, {                // 1319
        basePath: PROTOCOL + "//" + request.headers.host,              // 1320
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                             // 1325
        }) : ["en-US", "en"]                                           //
      });                                                              //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), WebSession.typeId, params, new Buffer(this.tabId, "hex")).session;
    }                                                                  //
                                                                       //
    return _callNewWebSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewApiSession = (function () {                  // 1183
    function _callNewApiSession(request, userInfo) {                   // 1333
      var _this4 = this;                                               //
                                                                       //
      var serializedParams = this.apiSessionParams;                    // 1334
      if (!serializedParams) {                                         // 1335
        throw new Meteor.Error(500, "Should have already computed apiSessionParams.");
      }                                                                //
                                                                       //
      // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon
      // calling newSession with an ApiSession._id.                    //
      // Eventually we'll remove this logic once we're sure apps have updated.
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), ApiSession.typeId, serializedParams, new Buffer(this.tabId, "hex")).then(function (session) {
        return session.session;                                        // 1346
      }, function (err) {                                              //
        return _this4._callNewWebSession(request, userInfo);           // 1348
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewApiSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewSession = (function () {                     // 1183
    function _callNewSession(request, viewInfo) {                      // 1352
      var userInfo = _.clone(this.userInfo);                           // 1353
      var _this = this;                                                // 1354
      var promise = inMeteor(function () {                             // 1355
        var vertex = undefined;                                        // 1356
        if (_this.apiToken) {                                          // 1357
          vertex = { token: _this.apiToken };                          // 1358
        } else {                                                       //
          // (_this.identityId might be null; this is fine)            //
          vertex = { grain: { _id: _this.grainId, identityId: _this.identityId } };
        }                                                              //
                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo);
        if (!permissions) {                                            // 1365
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
        }                                                              //
                                                                       //
        Sessions.update({                                              // 1369
          _id: _this.sessionId                                         // 1370
        }, {                                                           //
          $set: {                                                      // 1372
            viewInfo: viewInfo,                                        // 1373
            permissions: permissions                                   // 1374
          }                                                            //
        });                                                            //
                                                                       //
        return permissions;                                            // 1378
      });                                                              //
                                                                       //
      return promise.then(function (permissions) {                     // 1381
        userInfo.permissions = permissions;                            // 1382
                                                                       //
        var numBytes = Math.ceil(permissions.length / 8);              // 1384
        var buf = new Buffer(numBytes);                                // 1385
        for (var ii = 0; ii < numBytes; ++ii) {                        // 1386
          buf[ii] = 0;                                                 // 1387
        }                                                              //
                                                                       //
        for (var ii = 0; ii < permissions.length; ++ii) {              // 1390
          var byteNum = Math.floor(ii / 8);                            // 1391
          var bitNum = ii % 8;                                         // 1392
          if (permissions[ii]) {                                       // 1393
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                 // 1394
          }                                                            //
        }                                                              //
                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                      // 1398
                                                                       //
        if (_this.isApi) {                                             // 1400
          return _this._callNewApiSession(request, userInfo);          // 1401
        } else {                                                       //
          return _this._callNewWebSession(request, userInfo);          // 1403
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewSession;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.getSession = (function () {                          // 1183
    function getSession(request) {                                     // 1408
      var _this5 = this;                                               //
                                                                       //
      if (!this.session) {                                             // 1409
        (function () {                                                 //
          _this5.getConnection(); // make sure we're connected         // 1410
          var _this = _this5;                                          // 1411
          var promise = _this5.uiView.getViewInfo().then(function (viewInfo) {
            return inMeteor(function () {                              // 1413
              Grains.update(_this.grainId, { $set: { cachedViewInfo: viewInfo } });
            }).then(function () {                                      //
              return _this._callNewSession(request, viewInfo);         // 1416
            });                                                        //
          }, function (error) {                                        //
            if (error.kjType === "failed" || error.kjType === "unimplemented") {
              // Method not implemented.                               //
              // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built
              //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have
              //   to accept 'failed' here at least until the next API bump.
              return _this._callNewSession(request, {});               // 1424
            } else {                                                   //
              return Promise.reject(error);                            // 1426
            }                                                          //
          });                                                          //
          _this5.session = new Capnp.Capability(promise, WebSession);  // 1429
        })();                                                          //
      }                                                                //
                                                                       //
      return this.session;                                             // 1432
    }                                                                  //
                                                                       //
    return getSession;                                                 //
  })();                                                                //
                                                                       //
  Proxy.prototype.keepAlive = (function () {                           // 1183
    function keepAlive() {                                             // 1435
      this.getConnection();                                            // 1436
      return this.supervisor.keepAlive();                              // 1437
    }                                                                  //
                                                                       //
    return keepAlive;                                                  //
  })();                                                                //
                                                                       //
  Proxy.prototype.resetConnection = (function () {                     // 1183
    function resetConnection() {                                       // 1440
      if (this.session) {                                              // 1441
        this.session.close();                                          // 1442
        delete this.session;                                           // 1443
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1446
        this.uiView.close();                                           // 1447
        delete this.uiView;                                            // 1448
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1451
        this.supervisor.close();                                       // 1452
        delete this.supervisor;                                        // 1453
      }                                                                //
    }                                                                  //
                                                                       //
    return resetConnection;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.maybeRetryAfterError = (function () {                // 1183
    function maybeRetryAfterError(error, retryCount) {                 // 1457
      // If the error may be caused by the grain dying or a network failure, try to restart it,
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.
      // `retryCount` should be incremented for every successful retry as part of the same request;
      // we only want to retry once.                                   //
      var _this = this;                                                // 1462
      if (SandstormBackend.shouldRestartGrain(error, retryCount)) {    // 1463
        this.resetConnection();                                        // 1464
        return inMeteor(function () {                                  // 1465
          _this.supervisor = globalBackend.continueGrain(_this.grainId).supervisor;
        });                                                            //
      } else {                                                         //
        throw error;                                                   // 1469
      }                                                                //
    }                                                                  //
                                                                       //
    return maybeRetryAfterError;                                       //
  })();                                                                //
                                                                       //
  Proxy.prototype.doSessionInit = (function () {                       // 1183
    function doSessionInit(request, response, requestPath) {           // 1473
      // jscs:disable disallowQuotedKeysInObjects                      //
      var path = requestPath || "/";                                   // 1475
                                                                       //
      // Check that the path is relative (ie. starts with a /).        //
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {
        response.writeHead(400, "Invalid path supplied", { "Content-Type": "text/plain" });
        response.end("Invalid path supplied.");                        // 1481
        return;                                                        // 1482
      }                                                                //
                                                                       //
      // Set the session ID.                                           //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));
                                                                       //
      response.setHeader("Cache-Control", "no-cache, private");        // 1488
                                                                       //
      // Redirect to the app's root URL.                               //
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67
      response.writeHead(303, "See Other", { "Location": encodeURI(path) });
      response.end();                                                  // 1494
    }                                                                  //
                                                                       //
    return doSessionInit;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.makeContext = (function () {                         // 1183
    function makeContext(request, response) {                          // 1497
      // Parses the cookies from the request, checks that the session ID is present and valid, then
      // returns the request context which contains the other cookies.  Throws an exception if the
      // session ID is missing or invalid.                             //
                                                                       //
      var context = {};                                                // 1502
                                                                       //
      if (this.hostId) {                                               // 1504
        var parseResult = parseCookies(request);                       // 1505
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {
          throw new Meteor.Error(403, "Unauthorized");                 // 1507
        }                                                              //
                                                                       //
        if (parseResult.cookies.length > 0) {                          // 1510
          context.cookies = parseResult.cookies;                       // 1511
        }                                                              //
      } else {                                                         //
        // jscs:disable disallowEmptyBlocks                            //
        // This is an API request. Cookies are not supported.          //
      }                                                                //
                                                                       //
      context.accept = parseAcceptHeader(request);                     // 1518
                                                                       //
      context.eTagPrecondition = parsePreconditionHeader(request);     // 1520
                                                                       //
      context.additionalHeaders = [];                                  // 1522
      WebSession.Context.headerWhitelist.forEach(function (headerName) {
        if (request.headers[headerName]) {                             // 1524
          context.additionalHeaders.push({                             // 1525
            name: headerName,                                          // 1526
            value: request.headers[headerName]                         // 1527
          });                                                          //
        };                                                             //
      });                                                              //
                                                                       //
      var promise = new Promise(function (resolve, reject) {           // 1532
        response.resolveResponseStream = resolve;                      // 1533
        response.rejectResponseStream = reject;                        // 1534
      });                                                              //
                                                                       //
      context.responseStream = new Capnp.Capability(promise, ByteStream);
                                                                       //
      return context;                                                  // 1539
    }                                                                  //
                                                                       //
    return makeContext;                                                //
  })();                                                                //
                                                                       //
  Proxy.prototype.translateResponse = (function () {                   // 1183
    function translateResponse(rpcResponse, response, request) {       // 1542
      if (this.hostId) {                                               // 1543
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));
        }                                                              //
                                                                       //
        // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.
        // (2) Prevents the app from navigating the parent frame.      //
        // (3) Prevents the app from opening popups.                   //
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).
      } else {                                                         //
          // jscs:disable validateQuoteMarks                           //
          // This is an API request. Cookies are not supported.        //
                                                                       //
          // We need to make sure caches know that different bearer tokens get totally different results.
          response.setHeader('Vary', 'Authorization');                 // 1559
                                                                       //
          // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
          response.setHeader('Access-Control-Allow-Origin', '*');      // 1562
                                                                       //
          // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
          // in a browser context. This policy should thoroughly neuter it.
          response.setHeader('Content-Security-Policy', "default-src 'none'; sandbox");
        }                                                              //
                                                                       //
      // On first response, update the session to have hasLoaded=true  //
      this.setHasLoaded();                                             // 1570
                                                                       //
      // TODO(security): Set X-Content-Type-Options: nosniff?          //
                                                                       //
      if ('content' in rpcResponse) {                                  // 1574
        var content = rpcResponse.content;                             // 1575
        var code = successCodes[content.statusCode];                   // 1576
        if (!code) {                                                   // 1577
          throw new Error('Unknown status code: ', content.statusCode);
        }                                                              //
                                                                       //
        if (content.mimeType) {                                        // 1581
          response.setHeader('Content-Type', content.mimeType);        // 1582
        }                                                              //
                                                                       //
        if (content.encoding) {                                        // 1585
          response.setHeader('Content-Encoding', content.encoding);    // 1586
        }                                                              //
                                                                       //
        if (content.language) {                                        // 1589
          response.setHeader('Content-Language', content.language);    // 1590
        }                                                              //
                                                                       //
        if (content.eTag) {                                            // 1593
          response.setHeader('ETag', composeETag(content.eTag));       // 1594
        }                                                              //
                                                                       //
        if ('disposition' in content && 'download' in content.disposition) {
          response.setHeader('Content-Disposition', 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                              //
                                                                       //
        if ('stream' in content.body) {                                // 1602
          if (request.method === 'HEAD') {                             // 1603
            content.body.stream.close();                               // 1604
            response.rejectResponseStream(new Error('HEAD request; content doesn\'t matter.'));
          } else {                                                     //
            var _ret3 = (function () {                                 //
              var streamHandle = content.body.stream;                  // 1607
              response.writeHead(code.id, code.title);                 // 1608
              var promise = new Promise(function (resolve, reject) {   // 1609
                response.resolveResponseStream(new Capnp.Capability(new ResponseStream(response, streamHandle, resolve, reject), ByteStream));
              });                                                      //
              promise.streamHandle = streamHandle;                     // 1613
              return {                                                 // 1614
                v: promise                                             //
              };                                                       //
            })();                                                      //
                                                                       //
            if (typeof _ret3 === "object") return _ret3.v;             //
          }                                                            //
        } else {                                                       //
          response.rejectResponseStream(new Error('Response content body was not a stream.'));
                                                                       //
          if ('bytes' in content.body) {                               // 1620
            response.setHeader('Content-Length', content.body.bytes.length);
          } else {                                                     //
            throw new Error('Unknown content body type.');             // 1623
          }                                                            //
        }                                                              //
                                                                       //
        response.writeHead(code.id, code.title);                       // 1627
                                                                       //
        if ('bytes' in content.body && request.method !== 'HEAD') {    // 1629
          response.write(content.body.bytes);                          // 1630
        }                                                              //
                                                                       //
        response.end();                                                // 1633
      } else if ('noContent' in rpcResponse) {                         //
        var noContent = rpcResponse.noContent;                         // 1635
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];
        response.writeHead(noContentCode.id, noContentCode.title);     // 1637
        response.end();                                                // 1638
      } else if ('preconditionFailed' in rpcResponse) {                //
        var preconditionFailed = rpcResponse.preconditionFailed;       // 1640
        if (request.method === 'GET' && 'if-none-match' in request.headers) {
          if (preconditionFailed.matchingETag) {                       // 1642
            response.setHeader('ETag', composeETag(preconditionFailed.matchingETag));
          }                                                            //
                                                                       //
          response.writeHead(304, 'Not Modified');                     // 1646
        } else {                                                       //
          response.writeHead(412, 'Precondition Failed');              // 1648
        }                                                              //
                                                                       //
        response.end();                                                // 1651
      } else if ('redirect' in rpcResponse) {                          //
        var redirect = rpcResponse.redirect;                           // 1653
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];
        response.writeHead(redirectCode.id, redirectCode.title, {      // 1655
          'Location': redirect.location                                // 1656
        });                                                            //
        response.end();                                                // 1658
      } else if ('clientError' in rpcResponse) {                       //
        var clientError = rpcResponse.clientError;                     // 1660
        var errorCode = errorCodes[clientError.statusCode];            // 1661
        if (!errorCode) {                                              // 1662
          throw new Error('Unknown status code: ', clientError.statusCode);
        }                                                              //
                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {            // 1666
          'Content-Type': 'text/html'                                  // 1667
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1670
          if (clientError.descriptionHtml) {                           // 1671
            response.write(clientError.descriptionHtml);               // 1672
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>' + errorCode.id + ': ' + errorCode.title + '</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1680
      } else if ('serverError' in rpcResponse) {                       //
        response.writeHead(500, 'Internal Server Error', {             // 1682
          'Content-Type': 'text/html'                                  // 1683
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1686
          if (rpcResponse.serverError.descriptionHtml) {               // 1687
            response.write(rpcResponse.serverError.descriptionHtml);   // 1688
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>500: Internal Server Error</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1695
      } else {                                                         //
        throw new Error('Unknown HTTP response type:\n' + JSON.stringify(rpcResponse));
      }                                                                //
                                                                       //
      return Promise.resolve(undefined);                               // 1700
    }                                                                  //
                                                                       //
    return translateResponse;                                          //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequest = (function () {                       // 1183
    function handleRequest(request, data, response, retryCount) {      // 1703
      var _this = this;                                                // 1704
      return Promise.resolve(undefined).then(function () {             // 1705
        return _this.makeContext(request, response);                   // 1706
      }).then(function (context) {                                     //
        // jscs:disable requireDotNotation                             //
        // Send the RPC.                                               //
        var path = request.url.slice(1); // remove leading '/'         // 1710
        var session = _this.getSession(request);                       // 1711
                                                                       //
        var requestContent = function () {                             // 1713
          return {                                                     // 1714
            content: data,                                             // 1715
            encoding: request.headers['content-encoding'],             // 1716
            mimeType: request.headers['content-type']                  // 1717
          };                                                           //
        };                                                             //
                                                                       //
        var xmlContent = function () {                                 // 1721
          var type = request.headers['content-type'] || 'application/xml;charset=utf-8';
          var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);
          if (!match) {                                                // 1724
            response.writeHead(415, 'Unsupported media type.', {       // 1725
              'Content-Type': 'text/plain'                             // 1726
            });                                                        //
            response.end('expected XML request body');                 // 1728
            throw new Error('expected XML request body');              // 1729
          }                                                            //
                                                                       //
          var charset = match[2] || 'ISO-8859-1';                      // 1732
                                                                       //
          var encoding = request.headers['content-encoding'];          // 1734
          if (encoding && encoding !== 'identity') {                   // 1735
            if (encoding !== 'gzip') throw new Error('unknown Content-Encoding: ' + encoding);
            data = gunzipSync(data);                                   // 1737
          }                                                            //
                                                                       //
          return data.toString(charset.toLowerCase() === 'utf-8' ? 'utf8' : 'binary');
        };                                                             //
                                                                       //
        var propfindDepth = function () {                              // 1743
          var depth = request.headers['depth'];                        // 1744
          return depth === '0' ? 'zero' : depth === '1' ? 'one' : 'infinity';
        };                                                             //
                                                                       //
        var shallow = function () {                                    // 1750
          return request.headers['depth'] === '0';                     // 1751
        };                                                             //
                                                                       //
        var noOverwrite = function () {                                // 1754
          return (request.headers['overwrite'] || '').toLowerCase() === 'f';
        };                                                             //
                                                                       //
        var destination = function () {                                // 1758
          var result = request.headers['destination'];                 // 1759
          if (!result) throw new Error('missing destination');         // 1760
          return Url.parse(result).path.slice(1); // remove leading '/'
        };                                                             //
                                                                       //
        if (request.method === 'GET' || request.method === 'HEAD') {   // 1764
          return session.get(path, context, request.method === 'HEAD');
        } else if (request.method === 'POST') {                        //
          return session.post(path, requestContent(), context);        // 1767
        } else if (request.method === 'PUT') {                         //
          return session.put(path, requestContent(), context);         // 1769
        } else if (request.method === 'PATCH') {                       //
          return session.patch(path, requestContent(), context);       // 1771
        } else if (request.method === 'DELETE') {                      //
          return session["delete"](path, context);                     // 1773
        } else if (request.method === 'PROPFIND') {                    //
          return session.propfind(path, xmlContent(), propfindDepth(), context);
        } else if (request.method === 'PROPPATCH') {                   //
          return session.proppatch(path, xmlContent(), context);       // 1777
        } else if (request.method === 'MKCOL') {                       //
          return session.mkcol(path, requestContent(), context);       // 1779
        } else if (request.method === 'COPY') {                        //
          return session.copy(path, destination(), noOverwrite(), shallow(), context);
        } else if (request.method === 'MOVE') {                        //
          return session.move(path, destination(), noOverwrite(), context);
        } else if (request.method === 'LOCK') {                        //
          return session.lock(path, xmlContent(), shallow(), context);
        } else if (request.method === 'UNLOCK') {                      //
          return session.unlock(path, request.headers['lock-token'], context);
        } else if (request.method === 'ACL') {                         //
          return session.acl(path, xmlContent(), context);             // 1789
        } else if (request.method === 'REPORT') {                      //
          return session.report(path, requestContent(), context);      // 1791
        } else if (request.method === 'OPTIONS') {                     //
          return session.options(path, context).then(function (options) {
            var dav = [];                                              // 1794
            if (options.davClass1) dav.push('1');                      // 1795
            if (options.davClass2) dav.push('2');                      // 1796
            if (options.davClass3) dav.push('3');                      // 1797
            if (options.davExtensions) {                               // 1798
              options.davExtensions.forEach(function (token) {         // 1799
                if (token.match(/^([a-zA-Z0-9!#$%&'*+.^_`|~-]+|<[\x21-\x7E]*>)$/)) {
                  dav.push(token);                                     // 1801
                }                                                      //
              });                                                      //
            }                                                          //
                                                                       //
            if (dav.length > 0) {                                      // 1806
              response.setHeader("DAV", dav.join(", "));               // 1807
              response.setHeader("Access-Control-Expose-Headers", "DAV");
            }                                                          //
                                                                       //
            response.end();                                            // 1811
            // Return no response; we already handled everything.      //
          }, function (err) {                                          //
            if (err.kjType !== 'unimplemented') throw err;             // 1814
            response.end();                                            // 1815
            // Return no response; we already handled everything.      //
          });                                                          //
        } else {                                                       //
            throw new Error('Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.');
          }                                                            //
      }).then(function (rpcResponse) {                                 //
        if (rpcResponse !== undefined) {                               // 1822
          // Will be undefined for OPTIONS request.                    //
          return _this.translateResponse(rpcResponse, response, request);
        }                                                              //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleRequest(request, data, response, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequest;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequestStreaming = (function () {              // 1183
    function handleRequestStreaming(request, response, contentLength, retryCount) {
      var _this = this;                                                // 1833
      var context = this.makeContext(request, response);               // 1834
      var path = request.url.slice(1); // remove leading '/'           // 1835
      var session = this.getSession(request);                          // 1836
                                                                       //
      var mimeType = request.headers['content-type'] || 'application/octet-stream';
      var encoding = request.headers['content-encoding'];              // 1839
                                                                       //
      var requestStreamPromise = undefined;                            // 1841
      if (request.method === 'POST') {                                 // 1842
        requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);
      } else if (request.method === 'PUT') {                           //
        requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);
      } else {                                                         //
        throw new Error('Sandstorm only supports streaming POST and PUT requests.');
      }                                                                //
                                                                       //
      // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to
      //   handle the case where the app doesn't actually support streaming. We could pipeline while
      //   also buffering the data on the side in case we need it again later, but that's kind of
      //   complicated. We should fix the whole protocol to make streaming the standard.
      return requestStreamPromise.then(function (requestStreamResult) {
        var requestStream = requestStreamResult.stream;                // 1855
                                                                       //
        // Initialized when getResponse() returns, if the response is streaming.
        var downloadStreamHandle = undefined;                          // 1858
                                                                       //
        // Initialized if an upload-stream method throws.              //
        var uploadStreamError = undefined;                             // 1861
                                                                       //
        // We call `getResponse()` immediately so that the app can start streaming data down even while
        // data is still being streamed up. This theoretically allows apps to perform bidirectional
        // streaming, though probably very few actually do that.       //
        //                                                             //
        // Note that we need to be able to cancel `responsePromise` below, so it's important that it is
        // the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate variable.
        var responsePromise = requestStream.getResponse();             // 1869
                                                                       //
        var reportUploadStreamError = function (err) {                 // 1871
          // Called when an upload-stream method throws.               //
                                                                       //
          if (!uploadStreamError) {                                    // 1874
            uploadStreamError = err;                                   // 1875
                                                                       //
            // If we're still waiting on any response stuff, cancel it.
            responsePromise.cancel();                                  // 1878
            requestStream.close();                                     // 1879
            if (downloadStreamHandle) {                                // 1880
              downloadStreamHandle.close();                            // 1881
            }                                                          //
          }                                                            //
        };                                                             //
                                                                       //
        // If we have a Content-Length, pass it along to the app by calling `expectSize()`.
        if (contentLength !== undefined) {                             // 1887
          requestStream.expectSize(contentLength)["catch"](function (err) {
            // expectSize() is allowed to be unimplemented.            //
            if (err.kjType !== 'unimplemented') {                      // 1890
              reportUploadStreamError(err);                            // 1891
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        // Pipe the input stream to the app.                           //
        request.on('data', function (buf) {                            // 1897
          // TODO(soon): Only allow a small number of write()s to be in-flight at once,
          //   pausing the input stream if we hit that limit, so that we block the TCP socket all the
          //   way back to the source. May want to also coalesce small writes for this purpose.
          // TODO(security): The above problem may allow a DoS attack on the front-end.
          if (!uploadStreamError) requestStream.write(buf)["catch"](reportUploadStreamError);
        });                                                            //
                                                                       //
        request.on('end', function () {                                // 1905
          if (!uploadStreamError) requestStream.done()["catch"](reportUploadStreamError);
                                                                       //
          // We're all done making calls to requestStream.             //
          requestStream.close();                                       // 1909
        });                                                            //
                                                                       //
        request.on('close', function () {                              // 1912
          reportUploadStreamError(new Error('HTTP connection unexpectedly closed during request.'));
        });                                                            //
                                                                       //
        request.on('error', function (err) {                           // 1916
          reportUploadStreamError(err);                                // 1917
        });                                                            //
                                                                       //
        return responsePromise.then(function (rpcResponse) {           // 1920
          // Stop here if the upload stream has already failed.        //
          if (uploadStreamError) throw uploadStreamError;              // 1922
          var promise = _this.translateResponse(rpcResponse, response, request);
          downloadStreamHandle = promise.streamHandle;                 // 1924
          return promise;                                              // 1925
        });                                                            //
      }, function (err) {                                              //
        if (err.kjType === 'failed' && err.message.indexOf('not implemented') !== -1) {
          // Hack to work around old apps using an old version of Cap'n Proto, before the
          // 'unimplemented' exception type was introduced. :(         //
          // TODO(cleanup): When we transition to API version 2, we can move this into the
          //   compatibility layer.                                    //
          err.kjType = 'unimplemented';                                // 1933
        }                                                              //
                                                                       //
        if (SandstormBackend.shouldRestartGrain(err, 0)) {             // 1936
          // This is the kind of error that indicates we should retry. Note that we passed 0 for the
          // retry count above because we were just checking if this is a retriable error (vs. possibly
          // a method-not-implemented error); maybeRetryAfterError() will check again with the proper
          // retry count.                                              //
          return _this.maybeRetryAfterError(err, retryCount).then(function () {
            return _this.handleRequestStreaming(request, response, contentLength, retryCount + 1);
          });                                                          //
        } else if (err.kjType === 'unimplemented') {                   //
          // Streaming is not implemented. Fall back to non-streaming version.
          return readAll(request).then(function (data) {               // 1946
            return _this.handleRequest(request, data, response, 0);    // 1947
          });                                                          //
        } else {                                                       //
          throw err;                                                   // 1950
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequestStreaming;                                     //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleWebSocket = (function () {                     // 1183
    function handleWebSocket(request, socket, head, retryCount) {      // 1955
      var _this = this;                                                // 1956
                                                                       //
      return Promise.resolve(undefined).then(function () {             // 1958
        return _this.makeContext(request);                             // 1959
      }).then(function (context) {                                     //
        var path = request.url.slice(1); // remove leading '/'         // 1961
        var session = _this.getSession(request);                       // 1962
                                                                       //
        if (!('sec-websocket-key' in request.headers)) {               // 1964
          throw new Error('Missing Sec-WebSocket-Accept header.');     // 1965
        }                                                              //
                                                                       //
        var magic = request.headers['sec-websocket-key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        var acceptKey = Crypto.createHash('sha1').update(magic).digest('base64');
                                                                       //
        var protocols = [];                                            // 1971
        if ('sec-websocket-protocol' in request.headers) {             // 1972
          protocols = request.headers['sec-websocket-protocol'].split(',').map(function (s) {
            return s.trim();                                           // 1974
          });                                                          //
        }                                                              //
                                                                       //
        var receiver = new WebSocketReceiver(socket);                  // 1977
                                                                       //
        var promise = session.openWebSocket(path, context, protocols, receiver);
                                                                       //
        if (head.length > 0) {                                         // 1981
          promise.serverStream.sendBytes(head);                        // 1982
        }                                                              //
                                                                       //
        var socketIdx = _this.websocketCounter.toString();             // 1985
        _this.websockets[socketIdx] = socket;                          // 1986
        _this.websocketCounter += 1;                                   // 1987
        pumpWebSocket(socket, promise.serverStream, function () {      // 1988
          delete _this.websockets[socketIdx];                          // 1988
        });                                                            //
                                                                       //
        return promise.then(function (response) {                      // 1990
          var headers = ['HTTP/1.1 101 Switching Protocols', 'Upgrade: websocket', 'Connection: Upgrade', 'Sec-WebSocket-Accept: ' + acceptKey];
          if (response.protocol && response.protocol.length > 0) {     // 1997
            headers.push('Sec-WebSocket-Protocol: ' + response.protocol.join(', '));
          }                                                            //
                                                                       //
          headers.push('');                                            // 2001
          headers.push('');                                            // 2002
                                                                       //
          socket.write(headers.join('\r\n'));                          // 2004
          receiver.go();                                               // 2005
                                                                       //
          // Note:  At this point errors are out of our hands.         //
        });                                                            //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleWebSocket(request, socket, head, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleWebSocket;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.setHasLoaded = (function () {                        // 1183
    function setHasLoaded() {                                          // 2016
      var _this6 = this;                                               //
                                                                       //
      if (!this.hasLoaded) {                                           // 2017
        (function () {                                                 //
          _this6.hasLoaded = true;                                     // 2018
          var sessionId = _this6.sessionId;                            // 2019
          inMeteor(function () {                                       // 2020
            Sessions.update({ _id: sessionId }, { $set: { hasLoaded: true } });
          });                                                          //
        })();                                                          //
      }                                                                //
    }                                                                  //
                                                                       //
    return setHasLoaded;                                               //
  })();                                                                //
                                                                       //
  return Proxy;                                                        //
})();                                                                  //
                                                                       //
;                                                                      // 2025
                                                                       //
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;               // 2027
                                                                       //
var isRfc1918OrLocal = function (address) {                            // 2029
  if (Net.isIPv4(address)) {                                           // 2030
    quad = address.split('.').map(function (x) {                       // 2031
      return parseInt(x, 10);                                          // 2031
    });                                                                //
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                    //
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks
    return address === '::1' || address.lastIndexOf('fd', 0) === 0;    // 2037
  } else {                                                             //
    // Ignore things that are neither IPv4 nor IPv6                    //
    return false;                                                      // 2040
  }                                                                    //
};                                                                     //
                                                                       //
var quadToIntString = function (quad) {                                // 2044
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);
  return num.toString();                                               // 2049
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Session cookie management                                           //
                                                                       //
var parseCookies = function (request) {                                // 2055
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['cookie'];                              // 2057
                                                                       //
  var result = { cookies: [] };                                        // 2059
  if (header) {                                                        // 2060
    var reqCookies = header.split(';');                                // 2061
    for (var i in babelHelpers.sanitizeForInObject(reqCookies)) {      // 2062
      var reqCookie = reqCookies[i];                                   // 2063
      var equalsPos = reqCookie.indexOf('=');                          // 2064
      var cookie = undefined;                                          // 2065
      if (equalsPos === -1) {                                          // 2066
        cookie = { key: reqCookie.trim(), value: '' };                 // 2067
      } else {                                                         //
        cookie = { key: reqCookie.slice(0, equalsPos).trim(), value: reqCookie.slice(equalsPos + 1) };
      }                                                                //
                                                                       //
      if (cookie.key === 'sandstorm-sid') {                            // 2072
        if (result.sessionId) {                                        // 2073
          throw new Error('Multiple sandstorm session IDs?');          // 2074
        }                                                              //
                                                                       //
        result.sessionId = cookie.value;                               // 2077
      } else {                                                         //
        result.cookies.push(cookie);                                   // 2079
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2084
};                                                                     //
                                                                       //
var parsePreconditionHeader = function (request) {                     // 2087
  if (request.headers['if-match']) {                                   // 2088
    if (request.headers['if-match'].trim() === '*') {                  // 2089
      return { exists: null };                                         // 2090
    }                                                                  //
                                                                       //
    var matches = parseETagList(request.headers['if-match']);          // 2093
    if (matches.length > 0) {                                          // 2094
      return { matchesOneOf: matches };                                // 2095
    }                                                                  //
  }                                                                    //
                                                                       //
  if (request.headers['if-none-match']) {                              // 2099
    if (request.headers['if-none-match'].trim() === '*') {             // 2100
      return { doesntExist: null };                                    // 2101
    }                                                                  //
                                                                       //
    var noneMatches = parseETagList(request.headers['if-none-match']);
    if (noneMatches.length > 0) {                                      // 2105
      return { matchesNoneOf: noneMatches };                           // 2106
    }                                                                  //
  }                                                                    //
                                                                       //
  return { none: null };                                               // 2110
};                                                                     //
                                                                       //
var parseETagList = function (input) {                                 // 2113
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to
  // indicate that it is weak. We are parsing a list of comma-delimited etags.
                                                                       //
  input = input.trim();                                                // 2117
  var results = [];                                                    // 2118
                                                                       //
  while (input.length > 0) {                                           // 2120
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);    // 2121
    if (!match) throw new Meteor.Error(400, 'invalid etag');           // 2122
                                                                       //
    input = input.slice(match[0].length).trim();                       // 2124
    results.push({ weak: !!match[1], value: match[2].replace(/\\(.)/g, '$1') });
  }                                                                    //
                                                                       //
  return results;                                                      // 2128
};                                                                     //
                                                                       //
var composeETag = function (tag) {                                     // 2131
  var result = '"' + (tag.value || '').replace(/([\\"])/g, '\\$1') + '"';
  if (tag.weak) result = 'W/' + result;                                // 2133
  return result;                                                       // 2134
};                                                                     //
                                                                       //
var parseAcceptHeader = function (request) {                           // 2137
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['accept'];                              // 2139
                                                                       //
  var result = [];                                                     // 2141
  if (header) {                                                        // 2142
    var acceptList = header.split(',');                                // 2143
    for (var i in babelHelpers.sanitizeForInObject(acceptList)) {      // 2144
      var acceptStr = acceptList[i];                                   // 2145
      var tokensList = acceptStr.split(';');                           // 2146
                                                                       //
      var temp = { mimeType: tokensList[0].trim() };                   // 2148
                                                                       //
      var tokensListRest = tokensList.slice(1);                        // 2150
      for (var j in babelHelpers.sanitizeForInObject(tokensListRest)) {
        var token = tokensListRest[j];                                 // 2152
        var equalsPos = token.indexOf('=');                            // 2153
        if (equalsPos) {                                               // 2154
          var key = token.slice(0, equalsPos).trim();                  // 2155
          var value = token.slice(equalsPos + 1).trim();               // 2156
                                                                       //
          if (key === 'q') {                                           // 2158
            temp.qValue = +value;                                      // 2159
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      result.push(temp);                                               // 2164
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2168
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Regular HTTP request handling                                       //
                                                                       //
var readAll = function (stream) {                                      // 2174
  return new Promise(function (resolve, reject) {                      // 2175
    var buffers = [];                                                  // 2176
    stream.on('data', function (buf) {                                 // 2177
      buffers.push(buf);                                               // 2178
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 2181
      resolve(Buffer.concat(buffers));                                 // 2182
    });                                                                //
                                                                       //
    stream.on('error', reject);                                        // 2185
  });                                                                  //
};                                                                     //
                                                                       //
var makeSetCookieHeader = function (cookie) {                          // 2189
  var result = [cookie.name, '=', cookie.value];                       // 2190
                                                                       //
  if ('absolute' in cookie.expires) {                                  // 2192
    result.push('; Expires=');                                         // 2193
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());
  } else if ('relative' in cookie.expires) {                           //
    result.push('; Max-Age=' + cookie.expires.relative);               // 2196
  }                                                                    //
                                                                       //
  if (cookie.path) {                                                   // 2199
    result.push('; Path=' + cookie.path);                              // 2200
  }                                                                    //
                                                                       //
  if (cookie.httpOnly) {                                               // 2203
    result.push('; HttpOnly');                                         // 2204
  }                                                                    //
                                                                       //
  return result.join('');                                              // 2207
};                                                                     //
                                                                       //
// TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.
var successCodes = {                                                   // 2211
  ok: { id: 200, title: 'OK' },                                        // 2212
  created: { id: 201, title: 'Created' },                              // 2213
  accepted: { id: 202, title: 'Accepted' },                            // 2214
  multiStatus: { id: 207, title: 'Multi-Status' }                      // 2215
};                                                                     //
var noContentSuccessCodes = [                                          // 2217
// Indexed by shouldResetForm * 1                                      //
{ id: 204, title: 'No Content' }, { id: 205, title: 'Reset Content' }];
var redirectCodes = [                                                  // 2222
// Indexed by switchToGet * 2 + isPermanent                            //
{ id: 307, title: 'Temporary Redirect' }, { id: 308, title: 'Permanent Redirect' }, { id: 303, title: 'See Other' }, { id: 301, title: 'Moved Permanently' }];
var errorCodes = {                                                     // 2229
  badRequest: { id: 400, title: 'Bad Request' },                       // 2230
  forbidden: { id: 403, title: 'Forbidden' },                          // 2231
  notFound: { id: 404, title: 'Not Found' },                           // 2232
  methodNotAllowed: { id: 405, title: 'Method Not Allowed' },          // 2233
  notAcceptable: { id: 406, title: 'Not Acceptable' },                 // 2234
  conflict: { id: 409, title: 'Conflict' },                            // 2235
  gone: { id: 410, title: 'Gone' },                                    // 2236
  requestEntityTooLarge: { id: 413, title: 'Request Entity Too Large' },
  requestUriTooLong: { id: 414, title: 'Request-URI Too Long' },       // 2238
  unsupportedMediaType: { id: 415, title: 'Unsupported Media Type' },  // 2239
  imATeapot: { id: 418, title: 'I\'m a teapot' },                      // 2240
  unprocessableEntity: { id: 422, title: 'Unprocessable Entity' }      // 2241
};                                                                     //
                                                                       //
ResponseStream = (function () {                                        // 2244
  function ResponseStream(response, streamHandle, resolve, reject) {   // 2245
    babelHelpers.classCallCheck(this, ResponseStream);                 //
                                                                       //
    this.response = response;                                          // 2246
    this.streamHandle = streamHandle;                                  // 2247
    this.resolve = resolve;                                            // 2248
    this.reject = reject;                                              // 2249
    this.ended = false;                                                // 2250
  }                                                                    //
                                                                       //
  ResponseStream.prototype.write = (function () {                      // 2244
    function write(data) {                                             // 2253
      this.response.write(data);                                       // 2254
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.done = (function () {                       // 2244
    function done() {                                                  // 2257
      this.response.end();                                             // 2258
      this.streamHandle.close();                                       // 2259
      this.ended = true;                                               // 2260
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.close = (function () {                      // 2244
    function close() {                                                 // 2263
      if (this.ended) {                                                // 2264
        this.resolve();                                                // 2265
      } else {                                                         //
        this.streamHandle.close();                                     // 2267
        this.reject(new Error('done() was never called on outbound stream.'));
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return ResponseStream;                                               //
})();                                                                  //
                                                                       //
// TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.
var Zlib = Npm.require('zlib');                                        // 2274
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib);                  // 2275
                                                                       //
// -----------------------------------------------------------------------------
// WebSocket handling                                                  //
                                                                       //
WebSocketReceiver = (function () {                                     // 2280
  function WebSocketReceiver(socket) {                                 // 2281
    babelHelpers.classCallCheck(this, WebSocketReceiver);              //
                                                                       //
    this.socket = socket;                                              // 2282
    this.queue = [];                                                   // 2283
  }                                                                    //
                                                                       //
  WebSocketReceiver.prototype.go = (function () {                      // 2280
    function go() {                                                    // 2286
      for (var i in babelHelpers.sanitizeForInObject(this.queue)) {    // 2287
        this.socket.write(this.queue[i]);                              // 2288
      }                                                                //
                                                                       //
      this.queue = null;                                               // 2291
    }                                                                  //
                                                                       //
    return go;                                                         //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.sendBytes = (function () {               // 2280
    function sendBytes(message) {                                      // 2294
      // TODO(someday):  Flow control of some sort?                    //
      if (this.queue === null) {                                       // 2296
        this.socket.write(message);                                    // 2297
      } else {                                                         //
        this.queue.push(message);                                      // 2299
      }                                                                //
    }                                                                  //
                                                                       //
    return sendBytes;                                                  //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.close = (function () {                   // 2280
    function close() {                                                 // 2303
      this.socket.end();                                               // 2304
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return WebSocketReceiver;                                            //
})();                                                                  //
                                                                       //
pumpWebSocket = function (socket, rpcStream, destructor) {             // 2308
  socket.on('data', function (chunk) {                                 // 2309
    rpcStream.sendBytes(chunk)["catch"](function (err) {               // 2310
      if (err.kjType !== 'disconnected') {                             // 2311
        console.error('WebSocket sendBytes failed: ' + err.stack);     // 2312
      }                                                                //
                                                                       //
      socket.destroy();                                                // 2315
    });                                                                //
  });                                                                  //
                                                                       //
  socket.on('end', function (chunk) {                                  // 2319
    rpcStream.close();                                                 // 2320
  });                                                                  //
                                                                       //
  socket.on("close", function () {                                     // 2323
    destructor();                                                      // 2324
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// Debug log access                                                    //
                                                                       //
Meteor.publish('grainLog', function (grainId) {                        // 2331
  var _this7 = this;                                                   //
                                                                       //
  check(grainId, String);                                              // 2332
  var id = 0;                                                          // 2333
  var grain = Grains.findOne(grainId);                                 // 2334
  if (!grain || !this.userId || grain.userId !== this.userId) {        // 2335
    this.added('grainLog', id++, { text: 'Only the grain owner can view the debug log.' });
    this.ready();                                                      // 2337
    return;                                                            // 2338
  }                                                                    //
                                                                       //
  var connected = false;                                               // 2341
  var _this = this;                                                    // 2342
                                                                       //
  var receiver = {                                                     // 2344
    write: function (data) {                                           // 2345
      connected = true;                                                // 2346
      _this.added('grainLog', id++, { text: data.toString('utf8') });  // 2347
    },                                                                 //
                                                                       //
    close: function () {                                               // 2350
      if (connected) {                                                 // 2351
        _this.added('grainLog', id++, {                                // 2352
          text: '*** lost connection to grain (probably because it shut down) ***'
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  try {                                                                // 2359
    (function () {                                                     //
      var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {
        return supervisor.watchLog(8192, receiver);                    // 2361
      })).handle;                                                      //
      connected = true;                                                // 2363
      _this7.onStop(function () {                                      // 2364
        handle.close();                                                // 2365
      });                                                              //
    })();                                                              //
  } catch (err) {                                                      //
    if (!connected) {                                                  // 2368
      this.added('grainLog', id++, {                                   // 2369
        text: '*** couldn\'t connect to grain (' + err + ') ***'       // 2370
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 2376
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=proxy.js.map
