(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MeteorWrapperLdapjs = Package['typ:ldapjs'].MeteorWrapperLdapjs;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var ECMAScript = Package.ecmascript.ECMAScript;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var Future, LDAP_DEFAULTS, LDAP;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                             //
// packages/accounts-ldap/ldap_server.js                                                                       //
//                                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                               //
Future = Npm.require("fibers/future");                                                                         // 1
                                                                                                               //
// At a minimum, set up LDAP_DEFAULTS.url and .dn according to                                                 //
// your needs. url should appear as 'ldap://your.url.here'                                                     //
// dn should appear in normal ldap format of comma separated attribute=value                                   //
// e.g. 'uid=someuser,cn=users,dc=somevalue'                                                                   //
LDAP_DEFAULTS = {                                                                                              // 7
  url: false,                                                                                                  // 8
  port: "389",                                                                                                 // 9
  dn: false,                                                                                                   // 10
  searchDN: false,                                                                                             // 11
  searchCredentials: false,                                                                                    // 12
  createNewUser: true,                                                                                         // 13
  defaultDomain: false,                                                                                        // 14
  searchResultsProfileMap: false,                                                                              // 15
  base: null,                                                                                                  // 16
  search: "(objectclass=*)",                                                                                   // 17
  ldapsCertificate: false                                                                                      // 18
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 @class LDAP                                                                                                   //
 @constructor                                                                                                  //
 */                                                                                                            //
var LDAP = function (options) {                                                                                // 25
  // Set options                                                                                               //
  this.options = _.clone(LDAP_DEFAULTS);                                                                       // 27
                                                                                                               //
  // Because NPM ldapjs module has some binary builds,                                                         //
  // We had to create a wraper package for it and build for                                                    //
  // certain architectures. The package typ:ldap-js exports                                                    //
  // 'MeteorWrapperLdapjs' which is a wrapper for the npm module                                               //
  this.ldapjs = MeteorWrapperLdapjs;                                                                           // 33
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 * Attempt to bind (authenticate) ldap                                                                         //
 * and perform a dn search if specified                                                                        //
 *                                                                                                             //
 * @method ldapCheck                                                                                           //
 *                                                                                                             //
 * @param {Object} options  Object with username, ldapPass and overrides for LDAP_DEFAULTS object.             //
 * Additionally the searchBeforeBind parameter can be specified, which is used to search for the DN            //
 * if not provided.                                                                                            //
 */                                                                                                            //
LDAP.prototype.ldapCheck = function (db, options) {                                                            // 46
                                                                                                               //
  var _this = this;                                                                                            // 48
                                                                                                               //
  options = options || {};                                                                                     // 50
                                                                                                               //
  if (options.hasOwnProperty("username") && options.hasOwnProperty("ldapPass")) {                              // 52
    var _ret = (function () {                                                                                  //
      var explicitDnSelected = db.getLdapExplicitDnSelected();                                                 // 53
      if (explicitDnSelected) {                                                                                // 54
        _this.options.dn = db.getLdapDnPattern().replace("$USERNAME", options.username);                       // 55
      } else {                                                                                                 //
        _this.options.base = db.getLdapBase();                                                                 // 57
        _this.options.searchBeforeBind = {};                                                                   // 58
        _this.options.searchBeforeBind[db.getLdapSearchUsername()] = options.username;                         // 59
        _this.options.filter = db.getLdapFilter() || "(objectclass=*)";                                        // 60
        _this.options.searchBindDn = db.getLdapSearchBindDn();                                                 // 61
        _this.options.searchBindPassword = db.getLdapSearchBindPassword();                                     // 62
      }                                                                                                        //
                                                                                                               //
      var resolved = false;                                                                                    // 65
      var ldapAsyncFut = new Future();                                                                         // 66
                                                                                                               //
      // Create ldap client                                                                                    //
      var fullUrl = db.getLdapUrl();                                                                           // 69
      var client = null;                                                                                       // 70
                                                                                                               //
      var errorFunc = function (err) {                                                                         // 72
        if (err) {                                                                                             // 73
          if (resolved) return;                                                                                // 74
          resolved = true;                                                                                     // 75
          ldapAsyncFut["return"]({                                                                             // 76
            error: err                                                                                         // 77
          });                                                                                                  //
        }                                                                                                      //
      };                                                                                                       //
                                                                                                               //
      if (fullUrl.indexOf("ldaps://") === 0) {                                                                 // 82
        client = _this.ldapjs.createClient({                                                                   // 83
          url: fullUrl,                                                                                        // 84
          tlsOptions: {                                                                                        // 85
            ca: [_this.options.ldapsCertificate]                                                               // 86
          }                                                                                                    //
        }, errorFunc);                                                                                         //
      } else {                                                                                                 //
        client = _this.ldapjs.createClient({                                                                   // 90
          url: fullUrl                                                                                         // 91
        }, errorFunc);                                                                                         //
      }                                                                                                        //
                                                                                                               //
      client.on("error", errorFunc);                                                                           // 95
                                                                                                               //
      // Slide @xyz.whatever from username if it was passed in                                                 //
      // and replace it with the domain specified in defaults                                                  //
      var emailSliceIndex = options.username.indexOf("@");                                                     // 99
      var username = undefined;                                                                                // 100
      var domain = _this.options.defaultDomain;                                                                // 101
                                                                                                               //
      // If user appended email domain, strip it out                                                           //
      // And use the defaults.defaultDomain if set                                                             //
      if (emailSliceIndex !== -1) {                                                                            // 105
        username = options.username.substring(0, emailSliceIndex);                                             // 106
        domain = domain || options.username.substring(emailSliceIndex + 1, options.username.length);           // 107
      } else {                                                                                                 //
        username = options.username;                                                                           // 109
      }                                                                                                        //
                                                                                                               //
      // If DN is provided, use it to bind                                                                     //
      if (explicitDnSelected) {                                                                                // 113
        // Attempt to bind to ldap server with provided info                                                   //
        client.bind(_this.options.searchDN || _this.options.dn, _this.options.searchCredentials || options.ldapPass, function (err) {
          try {                                                                                                // 118
            if (err) {                                                                                         // 119
              // Bind failure, return error                                                                    //
              throw new Meteor.Error(err.code, err.message);                                                   // 121
            }                                                                                                  //
                                                                                                               //
            var handleSearchProfile = function (retObject, bindAfterSearch) {                                  // 124
              retObject.emptySearch = true;                                                                    // 125
                                                                                                               //
              // use dn if given, else use the base for the ldap search                                        //
              var searchBase = _this.options.dn || _this.options.base;                                         // 128
              var searchOptions = {                                                                            // 129
                scope: "sub",                                                                                  // 130
                sizeLimit: 1                                                                                   // 131
              };                                                                                               //
                                                                                                               //
              client.search(searchBase, searchOptions, function (err, res) {                                   // 134
                if (err) {                                                                                     // 135
                  if (resolved) return;                                                                        // 136
                  resolved = true;                                                                             // 137
                  ldapAsyncFut["return"]({                                                                     // 138
                    error: err                                                                                 // 139
                  });                                                                                          //
                  return;                                                                                      // 141
                }                                                                                              //
                                                                                                               //
                var found = false;                                                                             // 144
                res.on("searchEntry", function (entry) {                                                       // 145
                  found = true;                                                                                // 146
                  retObject.emptySearch = false;                                                               // 147
                  // Add entry results to return object                                                        //
                  retObject.searchResults = _.omit(entry.object, "userPassword");                              // 149
                                                                                                               //
                  if (bindAfterSearch) {                                                                       // 151
                    client.bind(retObject.searchResults.dn, options.ldapPass, function (err) {                 // 152
                      try {                                                                                    // 153
                        if (err) {                                                                             // 154
                          throw new Meteor.Error(err.code, err.message);                                       // 155
                        }                                                                                      //
                                                                                                               //
                        resolved = true;                                                                       // 158
                        ldapAsyncFut["return"](retObject);                                                     // 159
                      } catch (e) {                                                                            //
                        resolved = true;                                                                       // 161
                        ldapAsyncFut["return"]({                                                               // 162
                          error: e                                                                             // 163
                        });                                                                                    //
                      }                                                                                        //
                    });                                                                                        //
                  } else {                                                                                     //
                    resolved = true;                                                                           // 168
                    ldapAsyncFut["return"](retObject);                                                         // 169
                  }                                                                                            //
                });                                                                                            //
                                                                                                               //
                res.on("error", errorFunc);                                                                    // 173
                                                                                                               //
                res.on("end", function () {                                                                    // 175
                  if (!found) {                                                                                // 176
                    resolved = true;                                                                           // 177
                    ldapAsyncFut["return"]({                                                                   // 178
                      error: new Meteor.Error(500, "No user found")                                            // 179
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
            };                                                                                                 //
                                                                                                               //
            var retObject = {                                                                                  // 187
              username: username,                                                                              // 188
              searchResults: null,                                                                             // 189
              email: domain ? username + "@" + domain : false,                                                 // 190
              dn: _this.options.dn                                                                             // 191
            };                                                                                                 //
                                                                                                               //
            if (_this.options.searchDN) {                                                                      // 194
              handleSearchProfile(retObject, true);                                                            // 195
            } else {                                                                                           //
              handleSearchProfile(retObject, false);                                                           // 197
            }                                                                                                  //
          } catch (e) {                                                                                        //
            if (resolved) return;                                                                              // 200
            resolved = true;                                                                                   // 201
            ldapAsyncFut["return"]({                                                                           // 202
              error: e                                                                                         // 203
            });                                                                                                //
          }                                                                                                    //
        });                                                                                                    //
      }                                                                                                        //
      // DN not provided, search for DN and use result to bind                                                 //
      else {                                                                                                   //
          var _ret2 = (function () {                                                                           //
            if (_this.options.searchBindDn) {                                                                  // 210
              var _ret3 = (function () {                                                                       //
                var ldapBindFut = new Future();                                                                // 211
                client.bind(_this.options.searchBindDn, _this.options.searchBindPassword, function (err) {     // 212
                  if (err) {                                                                                   // 214
                    ldapBindFut["throw"](err);                                                                 // 215
                  } else {                                                                                     //
                    ldapBindFut["return"]();                                                                   // 217
                  }                                                                                            //
                });                                                                                            //
                                                                                                               //
                try {                                                                                          // 222
                  ldapBindFut.wait();                                                                          // 223
                } catch (err) {                                                                                //
                  return {                                                                                     // 225
                    v: {                                                                                       //
                      v: {                                                                                     //
                        v: { error: err }                                                                      //
                      }                                                                                        //
                    }                                                                                          //
                  };                                                                                           //
                }                                                                                              //
              })();                                                                                            //
                                                                                                               //
              if (typeof _ret3 === "object") return _ret3.v;                                                   //
            }                                                                                                  //
            // initialize result                                                                               //
            var retObject = {                                                                                  // 229
              username: username,                                                                              // 230
              email: domain ? username + "@" + domain : false,                                                 // 231
              emptySearch: true,                                                                               // 232
              searchResults: {}                                                                                // 233
            };                                                                                                 //
                                                                                                               //
            var filter = _this.options.filter;                                                                 // 236
            Object.keys(_this.options.searchBeforeBind).forEach(function (searchKey) {                         // 237
              filter = "&" + filter + "(" + searchKey + "=" + _this.options.searchBeforeBind[searchKey] + ")";
            });                                                                                                //
                                                                                                               //
            var searchOptions = {                                                                              // 241
              scope: "sub",                                                                                    // 242
              sizeLimit: 1,                                                                                    // 243
              filter: filter                                                                                   // 244
            };                                                                                                 //
                                                                                                               //
            // perform LDAP search to determine DN                                                             //
            client.search(_this.options.base, searchOptions, function (err, res) {                             // 248
              if (err) {                                                                                       // 249
                if (resolved) return;                                                                          // 250
                resolved = true;                                                                               // 251
                ldapAsyncFut["return"]({                                                                       // 252
                  error: err                                                                                   // 253
                });                                                                                            //
                return;                                                                                        // 255
              }                                                                                                //
                                                                                                               //
              retObject.emptySearch = true;                                                                    // 258
              res.on("searchEntry", function (entry) {                                                         // 259
                retObject.dn = entry.objectName;                                                               // 260
                retObject.username = retObject.dn;                                                             // 261
                retObject.emptySearch = false;                                                                 // 262
                                                                                                               //
                retObject.searchResults = _.omit(entry.object, "userPassword");                                // 264
                                                                                                               //
                // use the determined DN to bind                                                               //
                client.bind(entry.objectName, options.ldapPass, function (err) {                               // 267
                  try {                                                                                        // 268
                    if (err) {                                                                                 // 269
                      throw new Meteor.Error(err.code, err.message);                                           // 270
                    } else {                                                                                   //
                      resolved = true;                                                                         // 272
                      ldapAsyncFut["return"](retObject);                                                       // 273
                    }                                                                                          //
                  } catch (e) {                                                                                //
                    if (resolved) return;                                                                      // 277
                    resolved = true;                                                                           // 278
                    ldapAsyncFut["return"]({                                                                   // 279
                      error: e                                                                                 // 280
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
                                                                                                               //
              res.on("error", errorFunc);                                                                      // 286
                                                                                                               //
              // If no dn is found, return as is.                                                              //
              res.on("end", function (result) {                                                                // 289
                if (retObject.dn === undefined) {                                                              // 290
                  resolved = true;                                                                             // 291
                  ldapAsyncFut["return"](retObject);                                                           // 292
                }                                                                                              //
              });                                                                                              //
            });                                                                                                //
          })();                                                                                                //
                                                                                                               //
          if (typeof _ret2 === "object") return _ret2.v;                                                       //
        }                                                                                                      //
                                                                                                               //
      return {                                                                                                 // 298
        v: ldapAsyncFut.wait()                                                                                 //
      };                                                                                                       //
    })();                                                                                                      //
                                                                                                               //
    if (typeof _ret === "object") return _ret.v;                                                               //
  } else {                                                                                                     //
    throw new Meteor.Error(403, "Missing LDAP Auth Parameter");                                                // 301
  }                                                                                                            //
};                                                                                                             //
                                                                                                               //
// Register login handler with Meteor                                                                          //
// Here we create a new LDAP instance with options passed from                                                 //
// Meteor.loginWithLDAP on client side                                                                         //
// @param {Object} loginRequest will consist of username, ldapPass, ldap, and ldapOptions                      //
Accounts.registerLoginHandler("ldap", function (loginRequest) {                                                // 310
  // If 'ldap' isn't set in loginRequest object,                                                               //
  // then this isn't the proper handler (return undefined)                                                     //
  if (!loginRequest.ldap) {                                                                                    // 313
    return undefined;                                                                                          // 314
  }                                                                                                            //
                                                                                                               //
  if (!Accounts.identityServices.ldap.isEnabled()) {                                                           // 317
    throw new Meteor.Error(403, "LDAP service is disabled.");                                                  // 318
  }                                                                                                            //
                                                                                                               //
  // Instantiate LDAP with options                                                                             //
  var userOptions = loginRequest.ldapOptions || {};                                                            // 322
  var ldapObj = new LDAP(userOptions);                                                                         // 323
                                                                                                               //
  // Call ldapCheck and get response                                                                           //
  var ldapResponse = ldapObj.ldapCheck(this.connection.sandstormDb, loginRequest);                             // 326
                                                                                                               //
  if (ldapResponse.error) {                                                                                    // 328
    return {                                                                                                   // 329
      userId: null,                                                                                            // 330
      error: ldapResponse.error                                                                                // 331
    };                                                                                                         //
  } else if (ldapResponse.emptySearch) {                                                                       //
    return {                                                                                                   // 334
      userId: null,                                                                                            // 335
      error: new Meteor.Error(403, "User not found in LDAP")                                                   // 336
    };                                                                                                         //
  } else {                                                                                                     //
    // Set initial userId and token vals                                                                       //
    return Accounts.updateOrCreateUserFromExternalService("ldap", { id: ldapResponse.dn, rawAttrs: ldapResponse.searchResults }, {});
  }                                                                                                            //
});                                                                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-ldap'] = {
  LDAP: LDAP,
  LDAP_DEFAULTS: LDAP_DEFAULTS
};

})();

//# sourceMappingURL=accounts-ldap.js.map
